import numpy as np
import scipy.linalg
from peegy.processing.tools.template_generator.auditory_waveforms import eog_template
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data, filt_filt_data
from peegy.processing.tools.math_tools import get_local_maxima, get_local_minima
from peegy.processing.tools.filters.eegFiltering import bandpass_fir_win
from scipy import signal
import astropy.units as u
from tqdm import tqdm
import numba
from numba_progress import ProgressBar
from scipy.signal import correlate, windows
# from sklearn.neighbors import KernelDensity
import matplotlib.pyplot as plt


def rms(x: np.array = None):
    return np.sqrt(np.mean(x ** 2.0))


def crest_factor(x: np.array = None):
    return 20.0 * np.log10(np.max(np.abs(x)) / rms(x))


def unitary_step(x: np.array = None):
    return (x > 0).astype(float)


def generate_eog_template(eog_data: np.array = None,
                          fs: u.Quantity = None,
                          template_width: u.Quantity = 0.5 * u.s,
                          low_pass: u.Quantity = 20 * u.Hz,
                          crest_factor_threshold=10,
                          minimum_interval_width: u.Quantity = 0.075 * u.s,
                          eog_peak_width: u.Quantity = 0.02 * u.s,
                          n_iterations: int = 1,
                          kernel_bandwidth: float = 0.15,
                          use_initial_template=True
                          ):
    """
    This function generate an EOG template
    :param eog_data: eog data
    :param fs: sampling rate
    :param template_width: maximum width of the EOG template
    :param low_pass: lowpass frequency applied to smooth the eog template
    :param crest_factor_threshold: crest factor (in dB) used to detect and provide a first guess template from the data
    :param minimum_interval_width: minimum interval between eye blinks (only used for initial guess)
    :param eog_peak_width: empirical width of an EOG blink (only used for initial guess)
    :param n_iterations: number of iterations to improve the template estimation
    :param kernel_bandwidth: factor use to control the with of the Gaussian kernel in threshold detection
    :param use_initial_template: if true, a template eye blinking will be used as starting point to find events
    :return: template, peaks location
    """
    eog_data_to_fit = eog_data
    if low_pass is not None:
        _b = bandpass_fir_win(high_pass=None, low_pass=low_pass, fs=fs)
    _rms = rms(eog_data_to_fit)
    # we check if positive or negative dominated waveform by looking the number of events with crest factors
    # larger than crest_factor_threshold in dB
    _amp_threshold = 10 ** (crest_factor_threshold / 20) * _rms
    positive_peaks, _ = signal.find_peaks(
        (eog_data_to_fit * unitary_step(eog_data_to_fit - _amp_threshold)).squeeze(),
        prominence=_amp_threshold.to_value(eog_data_to_fit.unit),
        distance=int(fs * minimum_interval_width),
        width=int(fs * eog_peak_width))
    negative_peaks, _ = signal.find_peaks(
        (-eog_data_to_fit.value * unitary_step(-eog_data_to_fit - _amp_threshold)).squeeze(),
        prominence=_amp_threshold.to_value(eog_data_to_fit.unit),
        distance=int(fs * minimum_interval_width),
        width=int(fs * eog_peak_width))

    # first we generate a template by detecting peaks from the data
    template = np.zeros((int(template_width * fs), 1)) * u.dimensionless_unscaled
    _invert_waveform = False
    if use_initial_template:
        template, _ = eog_template(fs=fs, event_duration=template_width)
        # fit twice to find which polarity is most effective to fit data
        _idx_peak_1, _, _ = detect_blinks(eog_data=eog_data_to_fit,
                                          template=template.squeeze(),
                                          kernel_bandwidth=kernel_bandwidth)
        reconstructed_1 = reconstruct_eog(data=eog_data_to_fit.reshape(-1, 1),
                                          template=template,
                                          blink_positions=_idx_peak_1)
        _idx_peak_2, _, _ = detect_blinks(eog_data=eog_data_to_fit,
                                          template=-template.squeeze(),
                                          kernel_bandwidth=kernel_bandwidth)
        reconstructed_2 = reconstruct_eog(data=eog_data_to_fit.reshape(-1, 1),
                                          template=-template,
                                          blink_positions=_idx_peak_2)
        r1 = 0
        if _idx_peak_1.size:
            r1 = np.abs(np.corrcoef(eog_data_to_fit.squeeze(), reconstructed_1.squeeze())[0, 1])
        r2 = 0
        if _idx_peak_2.size:
            r2 = np.abs(np.corrcoef(-eog_data_to_fit.squeeze(), reconstructed_2.squeeze())[0, 1])
        if r1 > r2:
            blink_positions = _idx_peak_1
        else:
            blink_positions = _idx_peak_2
            _invert_waveform = True
        if _invert_waveform:
            template = template * -1
    else:
        if positive_peaks.size and negative_peaks.size:
            _invert_waveform = positive_peaks.size < negative_peaks.size
        elif positive_peaks.size:
            _invert_waveform = False
        elif negative_peaks.size:
            _invert_waveform = True
        _sign = (-1) ** _invert_waveform
        _idx_peak, _ = signal.find_peaks(
            (_sign * eog_data_to_fit * unitary_step(_sign * eog_data_to_fit - _amp_threshold)).squeeze(),
            distance=int(fs * minimum_interval_width),
            prominence=_amp_threshold.to_value(eog_data_to_fit.unit),
            width=int(eog_peak_width * fs))
        blink_positions = np.array([])
        if _idx_peak.size:
            blink_positions = _idx_peak - int(template_width * fs) // 2
            blink_positions = blink_positions[
                np.logical_and(blink_positions >= 0,
                               blink_positions < eog_data_to_fit.shape[0] - np.round(template_width * fs) - 1)]
            template = generate_eog_template_from_peaks(blink_pos=blink_positions,
                                                        eog_data=eog_data_to_fit,
                                                        fs=fs,
                                                        template_width=template_width,
                                                        b=_b)
    z = np.array([])
    if blink_positions.size:
        # template = generate_eog_template_from_peaks(blink_pos=blink_positions,
        #                                             eog_data=eog_data_to_fit,
        #                                             fs=fs,
        #                                             template_width=template_width,
        #                                             b=_b)
        for _i in tqdm(np.arange(n_iterations), desc='EOG template iteration'):
            blink_positions, z, threshold = detect_blinks(eog_data=eog_data_to_fit,
                                                          template=template.squeeze(),
                                                          kernel_bandwidth=kernel_bandwidth)
            if blink_positions.size:
                template = generate_eog_template_from_peaks(blink_pos=blink_positions,
                                                            eog_data=eog_data_to_fit,
                                                            fs=fs,
                                                            template_width=template_width,
                                                            b=_b)
            print('Iteration {:}; {:} blinks detected; threshold {:}'.format(
                _i,
                blink_positions.size,
                threshold
                ))
    # print final correlation
    reconstructed_final = reconstruct_eog(data=eog_data_to_fit.reshape(-1, 1), template=template,
                                          blink_positions=blink_positions)
    if np.std(reconstructed_final.squeeze()) > 0:
        _score = np.abs(np.corrcoef(eog_data_to_fit.squeeze(), reconstructed_final.squeeze())[0, 1])
        print('Correlation between reconstructed and target EOG data: {:}'.format(_score))
    else:
        print('No EOG events were found')
    return template, blink_positions, z


def generate_eog_template_from_peaks(
        eog_data: np.array = None,
        blink_pos: np.array = None,
        fs: u.Quantity = None,
        template_width: u.Quantity = 0.5 * u.s,
        b: np.array = None,
        fade_duration: u.Quantity = 0.2 * u.s
):
    """
    This function will generate a template EOG based on the peaks position provided in idx_peak.
    :param eog_data: eog data
    :param blink_pos: location of peaks (in samples)
    :param fs: sampling rate
    :param b: filter coefficients
    :param template_width: width of the EOG template
    :param fade_duration: duration of the linear fade-in and fade-out applied to template
    :return: a template generated by weighted average of data centred on the peaks with a width defined by

    template_width.
    """
    template_samples = int(fs * template_width)
    h = np.zeros((template_samples, blink_pos.shape[0])) * eog_data.unit
    _index = np.arange(template_samples)
    for _i, _ev_pos in enumerate(blink_pos):
        _current_data = eog_data[_index + _ev_pos].squeeze()
        h[:, _i] = _current_data - np.mean(_current_data, axis=0)
    h = np.sort(h, axis=1)
    i1 = round(blink_pos.shape[0] * 0.25)
    i3 = round(blink_pos.shape[0] * 0.75)
    h = h[:, i1: i3]  # Amplitudes between perc. 25 - 75
    window = windows.hamming(i3 - i1)
    window = window / sum(window)  # normalization
    template = h.dot(window)[:, None]
    if b is not None:
        template = filt_data(data=template, b=b)
    # apply window to have smooth onset and offset
    _window = np.ones((template.size, 1))
    nw = int(np.minimum(np.maximum(fade_duration * fs, 0), template.shape[0] // 2))
    _window[0: nw] = (np.arange(0, nw) / (nw - 1)).reshape(-1, 1)
    _window[_window.size - nw::] = _window[0: nw][::-1]
    template = template * _window
    # normalize
    template = template * template.unit / np.sqrt(np.sum(template ** 2))

    return template


@numba.jit(nogil=True, nopython=True)
def get_cross_correlation_norm(data: np.array = None,
                               template: np.array = None,
                               progress_proxy: ProgressBar = None) -> np.array:
    """
    This function computes the normalized correlation between two different size numpy arrays.
    The template is assumed to be shorter than the full data.
    The output will return the correlation in which case, if the template exactly matches a portion of the data,
    regardless amplitude differences, the value at that point will be 1 or -1 (if negatively correlated).
    :param data: numpy array with full data set to find matches with template.
    :param template: numpy array containing the target template to be found in the data
    :param progress_proxy: ProgressBar object to print progress of the filter
    :return: normalized correlation function
    """
    assert data.shape[0] >= template.shape[0], r'data input length must be greater or equal than template'
    norm = np.zeros(data.shape)
    # get variance of template
    n2 = np.sum(template ** 2)
    for _i in np.arange(data.shape[0]):
        _ini = _i
        _end = np.minimum(_i + template.shape[0], data.shape[0])
        x1 = data[_ini: _end]
        # get variance of each segment in the data
        norm[_i] = np.sum(x1 ** 2)
        progress_proxy.update(1)
    # compute correlation norm as sqrt(var_data * variance_template)
    norm = np.sqrt(norm * n2)
    return norm


def match_filter(data: np.array = None,
                 template: np.array = None) -> np.array:
    """
    This function computes the normalized correlation between two different size numpy arrays.
    The template is assumed to be shorter than the full data.
    The output will return the correlation in which case, if the template exactly matches a portion of the data,
    regardless amplitude differences, the value at that point will be 1 or -1 (if negatively correlated).
    :param data: numpy array with full data set to find matches with template.
    :param template: numpy array containing the target template to be found in the data
    :return: normalized correlation function
    """
    assert data.shape[0] >= template.shape[0], r'data input length must be greater or equal than template'
    num_iterations = data.shape[0] - template.shape[0]
    # compute cross-correlation and resize it
    corr = correlate(data, template, 'full')
    corr = corr[template.shape[0] - 1: data.shape[0] + template.shape[0]]
    # estimate the norm of the cross-correlation
    with ProgressBar(total=num_iterations, desc='Detecting peaks via template correlation') as progress:
        norm = get_cross_correlation_norm(data=data, template=template, progress_proxy=progress)
    norm_xcor = corr / norm
    return norm_xcor


def detect_blinks_2(eog_data: np.array = None,
                    template: np.array = None,
                    threshold: float = 0.8,
                    ):
    """
    This function will use a match filter based on normalized cross correlation to detect the position at which the
    input template matches best.
    :param eog_data: numpy array with eog data
    :param template: template generated from the eog data (see generate_eog_template)
    :param threshold: the correlation threshold to detect peaks
    :return: the time of the detected event (as unitary dirac pulses) at which the template should be located
    """

    x_corr = match_filter(data=eog_data.value.squeeze(),
                          template=template.value.squeeze())
    dirac_peaks, _ = signal.find_peaks(np.abs(x_corr),
                                       prominence=threshold,
                                       width=1)
    return dirac_peaks


def detect_blinks(eog_data: np.array = None,
                  template: np.array = None,
                  kernel_bandwidth: float = 0.15
                  ):
    """
    This function will use a match filter based on normalized cross correlation to detect the position at which the
    input template matches best.
    :param eog_data: numpy array with eog data
    :param template: template generated from the eog data (see generate_eog_template)
    :param kernel_bandwidth: factor use to control the with of the Gaussian kernel in threshold detection
    :return: the time of the detected event (as unitary dirac pulses) at which the template should be located and
    correlation function between eeg_data and template
    """
    z = filt_data(data=eog_data, b=template.value[::-1], mode='valid', onset_padding=False).value
    _maxima, _idx_maxima = get_local_maxima(z.squeeze())
    threshold = detect_threshold(_maxima, kernel_bandwidth=kernel_bandwidth, show_histogram=False)
    blinks_pos = blink_event_positions(local_maxima=_maxima,
                                       idx_local_maxima=_idx_maxima,
                                       threshold=threshold,
                                       n_samples=eog_data.shape[0],
                                       impulse_response_length=template.shape[0])
    return blinks_pos, z, threshold


def histc(x, bins):
    map_to_bins = np.digitize(x, bins)  # Get indices of the bins to which each value in input array belongs.
    res = np.zeros(bins.shape)
    for el in map_to_bins:
        res[el - 1] += 1  # Increment appropriate bin.
    return res


def detect_threshold(peaks: np.array,
                     kernel_bandwidth: float = 0.15,
                     show_histogram=False):
    q3, q1 = np.percentile(peaks, [75, 25])
    iqr = q3 - q1
    bw = iqr * kernel_bandwidth
    # kde = KernelDensity(bandwidth=bw, kernel='gaussian')
    # kde.fit(peaks[:, None])
    x_d = np.linspace(peaks.min() - 5 * bw,
                      peaks.max() + 5 * bw, 1000).reshape(-1, 1)

    da = x_d[1] - x_d[0]  # Step in the  axis  of  amplitudes
    # Gaussian kernel
    kernel = np.exp(-(np.arange(-4.0 * bw, 4.0 * bw, da) ** 2 / (2 * bw ** 2))) * u.dimensionless_unscaled
    kernel = kernel / np.sum(kernel)
    _histogram = histc(peaks, x_d.squeeze()) * u.dimensionless_unscaled
    _pdf = filt_filt_data(data=_histogram.reshape(-1, 1), b=kernel, onset_padding=False)
    # _pdf = np.exp(kde.score_samples(x_d))
    _minima, _minima_idx = get_local_minima(_pdf.squeeze())
    _max_idx = _pdf.argmax()
    threshold = None
    if _minima_idx[_minima_idx > _max_idx].size:
        threshold_idx = _minima_idx[_minima_idx > _max_idx][0]
        threshold = float(x_d[threshold_idx])
    if show_histogram:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(x_d, _histogram)
        ax.plot(x_d[threshold_idx], _pdf[threshold_idx], 'o')
        ax.fill(x_d, _pdf, fc="#AAAAFF")
        fig.show()
    return threshold


def blink_event_positions(local_maxima: np.array = None,
                          idx_local_maxima: np.array = None,
                          threshold: float = None,
                          impulse_response_length: int = None,
                          n_samples: int = None):
    _position = idx_local_maxima[local_maxima > threshold]
    # lag compensation
    _position = _position - (impulse_response_length - 1)
    _position = _position[np.logical_and(_position >= 0, _position < n_samples - impulse_response_length - 1)]
    return _position


def reconstruct_eog(data: np.array = None,
                    template: np.array = np.array([]),
                    blink_positions: np.array = np.array([])):
    """
    This function will reconstruct an EOG artefact by convolving the template (scaled to maximize similarity with input
    data) with unitary impulses located at the event_times
    :param data: data where EOG will be based from.
    :param template: an estimation of the EOG artefact
    :param blink_positions: location of the EOG artefacts
    :return: reconstructed EOG scaled to match the EOG in the input data
    """
    reconstructed = np.zeros(data.shape)
    if template.size and blink_positions.size:
        k = blink_positions.shape[0]
        rh = np.zeros((k, k))
        n_template = template.shape[0]
        auto_corr_h = correlate(template, template, 'full')
        auto_corr_h = auto_corr_h[(n_template - 1)::]
        for k_1 in range(k):
            for k_2 in range(k):
                dt = np.round(np.abs(blink_positions[k_2] - blink_positions[k_1]))
                if dt < n_template:
                    rh[k_1, k_2] = auto_corr_h[dt]
        inv_rh = scipy.linalg.pinv(rh)
        convolution_matrix = np.zeros(data.shape)
        template = template.squeeze()
        for _ch in tqdm(np.arange(data.shape[1]), desc='Computing EOG amplitudes'):
            z_channel = correlate(data[:, _ch], template, 'full')
            _positions = np.minimum(blink_positions + n_template, z_channel.shape[0] - 1)
            z_channel_amps = z_channel[_positions]  # Amplitudes for that channel
            amplitudes_ch = inv_rh.dot(z_channel_amps)
            convolution_matrix[blink_positions, _ch] = amplitudes_ch.squeeze()
        # convolve source with dirac train to generate blinks signal
        for _ch_idx in tqdm(np.arange(0, convolution_matrix.shape[1]), desc='Reconstructing EOG artefacts'):
            reconstructed[:, _ch_idx] = filt_data(
                data=template[:, None],
                b=convolution_matrix[:, _ch_idx].reshape(-1, 1).flatten(),
                mode='full', onset_padding=False)[0: data.shape[0]].flatten()
    return reconstructed





class TestType(object):
    hotelling_t2_time = 'hotelling_t2_time'  # hotelling-t2 test in the time-domain
    hotelling_t2_freq = 'hotelling_t2_freq'  # hotelling-t2 test in the frequency-domain
    f_test_freq = 'f_test_freq'  # f-ratio in frequency domain
    f_test_time = 'f_test_time'  # f-ratio using multiple points in the time-domain
    rayleigh_test = 'rayleigh_test'  # phase-locking value
    covariance = 'covariance'  # covariance
    bootstrap = 'bootstrap'

    def get_available_methods(self):
        members = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")]
        return members


class HotellingTSquareFrequencyTest(object):
    def __init__(self,
                 data_source: str = None,
                 test_name='HT2',
                 frequency_tested=None,
                 df_1=None,
                 df_2=None,
                 t_square=None,
                 f=None,
                 p_value=None,
                 n_epochs=None,
                 spectral_magnitude=None,
                 spectral_phase=None,
                 rn=None,
                 snr=None,
                 snr_db=None,
                 f_critic=None,
                 channel=None,
                 ini_time: float = None,
                 end_time: float = None
                 ):
        self.data_source = data_source
        self.test_name = test_name
        self.frequency_tested = frequency_tested
        self.df_1 = df_1
        self.df_2 = df_2
        self.t_square = t_square
        self.f = f
        self.p_value = p_value
        self.spectral_magnitude = spectral_magnitude
        self.spectral_phase = spectral_phase
        self.rn = rn
        self.n_epochs = n_epochs
        self.snr = snr
        self.snr_db = snr_db
        self.f_critic = f_critic
        self.channel = channel
        self.ini_time = ini_time
        self.end_time = end_time


class HotellingTSquareTest(object):
    def __init__(self,
                 test_name='HT2',
                 df_1=None,
                 df_2=None,
                 t_square=None,
                 f=None,
                 f_critic=None,
                 p_value=None,
                 mean_amplitude=None,
                 mean_phase=None,
                 rn=None,
                 n_epochs=None,
                 snr=None,
                 snr_db=None,
                 snr_critic_db=None,
                 snr_critic=None,
                 channel=None,
                 frequency_tested=None,
                 ini_time: float = None,
                 end_time: float = None,
                 **kwargs
                 ):
        self.test_name = test_name
        self.df_1 = df_1
        self.df_2 = df_2
        self.t_square = t_square
        self.f = f
        self.f_critic = f_critic
        self.p_value = p_value
        self.mean_amplitude = mean_amplitude
        self.mean_phase = mean_phase
        self.rn = rn
        self.n_epochs = n_epochs
        self.snr = snr
        self.snr_db = snr_db
        self.snr_critic_db = snr_critic_db
        self.snr_critic = snr_critic
        self.channel = channel
        self.frequency_tested = frequency_tested
        self.ini_time = ini_time
        self.end_time = end_time
        for _item, _value in kwargs.items():
            setattr(self, _item, _value)


class FrequencyFTest(object):
    def __init__(self,
                 test_name='F-test',
                 frequency_tested=None,
                 df_1=None,
                 df_2=None,
                 f=None,
                 p_value=None,
                 spectral_magnitude=None,
                 spectral_phase=None,
                 rn=None,
                 snr=None,
                 snr_db=None,
                 f_critic=None,
                 channel=None,
                 ini_time: float = None,
                 end_time: float = None,
                 ):
        self.test_name = test_name
        self.frequency_tested = frequency_tested
        self.df_1 = df_1
        self.df_2 = df_2
        self.f = f
        self.p_value = p_value
        self.spectral_magnitude = spectral_magnitude
        self.spectral_phase = spectral_phase
        self.rn = rn
        self.snr = snr
        self.snr_db = snr_db
        self.f_critic = f_critic
        self.channel = channel
        self.ini_time = ini_time
        self.end_time = end_time


class PhaseLockingValueTest(object):
    def __init__(self,
                 test_name='rayleigh_test',
                 plv=None,
                 df_1=None,
                 z_value=None,
                 z_critic=None,
                 p_value=None,
                 mean_phase=None,
                 channel=None,
                 frequency_tested=None,
                 ini_time: float = None,
                 end_time: float = None,
                 rn=None):
        self.test_name = test_name
        self.plv = plv
        self.df_1 = df_1
        self.z_value = z_value
        self.z_critic = z_critic
        self.p_value = p_value
        self.mean_phase = mean_phase
        self.channel = channel
        self.frequency_tested = frequency_tested
        self.rn = rn
        self.ini_time = ini_time
        self.end_time = end_time


class FmpTest(object):
    def __init__(self,
                 test_name='Fmp',
                 label: str = None,
                 df_1: float = None,
                 df_2: float = None,
                 f: float = None,
                 f_critic: float = None,
                 p_value: float = None,
                 rn: float = None,
                 snr: float = None,
                 ini_time: float = None,
                 end_time: float = None,
                 n_epochs: int = None,
                 channel: str = None):
        self.test_name = test_name
        self.label = label
        self.df_1 = df_1
        self.df_2 = df_2
        self.f = f
        self.f_critic = f_critic
        self.p_value = p_value
        self.rn = rn
        self.snr = snr
        self.ini_time = ini_time
        self.end_time = end_time
        self.n_epochs = n_epochs
        self.channel = channel



import astropy.units as u
from scipy import signal
import matplotlib.pyplot as plt
import numpy as np
# from ssqueezepy import Wavelet
# import pywt


reference_frequency = 40 * u.Hz
half_width = 2 / reference_frequency
fs = 1000 * u.Hz

# fwhm = (4 / reference_frequency)
# b = 2 * (fwhm / (2 * np.sqrt(2 * np.log(2)))) ** 2.0
# c = reference_frequency
# morlet = 'cmor{:}-{:}'.format(b.value, c.value)
# wavelet = pywt.ContinuousWavelet(morlet)
# [wav, time] = wavelet.wavefun(level=16)
# scale = pywt.frequency2scale(wavelet=wavelet, freq=reference_frequency/fs.value, precision=16).value
# print(pywt.scale2frequency(wavelet, [scale], precision=16) / (1/fs))
# freq = np.arange(0, wav.size) * 1 / np.mean(np.diff(time)) / wav.size
# yfft = np.fft.fft(wav, axis=0)
#
# plt.figure()
# plt.plot(freq, np.abs(yfft))
# plt.plot(reference_frequency.value, np.abs(yfft).max(), 'o')
# plt.xlim([0, 2 * reference_frequency.value])
# plt.xlabel('Frequency [Hz]')
# plt.show()
#
# n_wavelet = wav / np.max(np.abs(wav))
# _ini = np.argwhere(np.abs(n_wavelet) >= 0.5)[0]
# _end = np.argwhere(np.abs(n_wavelet) >= 0.5)[-1]
# full_width_half_max = (time[_end] - time[_ini])
# plt.figure()
# plt.plot(time, np.abs(n_wavelet))
# plt.plot(time[_ini], np.abs(n_wavelet)[_ini], 'o')
# plt.plot(time[_end], np.abs(n_wavelet)[_end], 'o')
# plt.plot(time, np.real(n_wavelet))
# plt.xlabel('Time [s]')
# plt.show()

# M is just for plot. We use CWT for convolution
M = 1000
time = np.arange(M) / fs
s = (half_width.to(u.s) * fs)
w = reference_frequency * (2 * s * np.pi) / fs
wavelet = signal.morlet2(M, s, w)
n_wavelet = wavelet / np.max(np.abs(wavelet))
_ini = np.argwhere(np.abs(n_wavelet) >= 0.5)[0]
_end = np.argwhere(np.abs(n_wavelet) >= 0.5)[-1]
full_width_half_max = (time[_end] - time[_ini])

plt.plot(time, n_wavelet)
plt.plot(time, np.abs(n_wavelet))
plt.plot(time, np.real(n_wavelet))
plt.plot(time[_ini], np.abs(n_wavelet)[_ini], 'o')
plt.plot(time[_end], np.abs(n_wavelet)[_end], 'o')
plt.show()


# define test stimuli and reference signal
duration = 1 * u.s
time = np.arange(0, int(duration * fs)) / fs
ini_phase = np.pi / 2 * u.rad
# reference is cosine convolved with wavelet so that phase includes the filter response
reference_signal = (np.cos(2 * np.pi * reference_frequency * u.rad * time[:, None] - ini_phase) +
                    1j * np.sin(2 * np.pi * reference_frequency * u.rad * time[:, None] - ini_phase))
reference_signal = reference_signal / np.abs(reference_signal)
noise_amp = 0.1
fig1, axs = plt.subplots(2)
# now we estimate the phase changes for a range of phase transitions at the mid-point of the time axis
for phase in np.linspace(0, np.pi, 10) * u.rad:
    # generate test signal
    data = (np.cos(2 * np.pi * u.rad * reference_frequency * time - ini_phase)[:, None] +
            noise_amp * np.random.randn(time.size, 1))
    # generate phase transition
    data[data.shape[0] // 10::] = (np.cos(2 * np.pi * u.rad * reference_frequency * time[data.shape[0] // 10::] -
                                          ini_phase + phase)[:, None] +
                                   noise_amp * np.random.randn(time[data.shape[0] // 10::].size, 1))

    data[data.shape[0] // 4::] = (np.cos(2 * np.pi * u.rad * reference_frequency * time[data.shape[0] // 4::] -
                                         ini_phase + 2 * phase)[:, None] +
                                  noise_amp * np.random.randn(time[data.shape[0] // 4::].size, 1))

    # convolve signal with complex-wavelet
    y_filter = np.zeros(data.shape, dtype=complex)
    ####
    # wavelet = Wavelet('morlet')
    # Wx, scales = cwt(data.value.T, wavelet)
    # freqs_cwt = scale_to_freq(scales, wavelet, y_filter.shape[0], fs=fs.value)
    # _ifreq = np.argwhere(freqs_cwt == reference_frequency.value)
    # y_filter = Wx.T[:, _ifreq[0], ...]
    ####
    if data.ndim == 1:
        y_filter = signal.cwt(data.value, signal.morlet2, widths=[s], w=w)
        # y_filter, _ = pywt.cwt(data.value,
        #                        wavelet=wavelet,
        #                        scales=[scale],
        #                        sampling_period=1 / fs.value)
        y_filter = y_filter[:, None, None]
    if data.ndim == 2:
        for _i in range(y_filter.shape[1]):
            y_filter[:, _i] = signal.cwt(data[:, _i].value, signal.morlet2,
                                         widths=[s], w=w)
            # y_filter[:, _i], _ = pywt.cwt(data[:, _i].value,
            #                               wavelet=wavelet,
            #                               scales=[scale],
            #                               sampling_period=1 / fs.value)
            y_filter = y_filter[:, None]
    if data.ndim == 3:
        for _i in range(y_filter.shape[1]):
            for _j in range(y_filter.shape[2]):
                y_filter[:, _i, _j] = signal.cwt(data[:, _i, _j].value, signal.morlet2,
                                                 widths=[s], w=w)
                # y_filter[:, _i, _j], _ = pywt.cwt(data[:, _i, _j].value,
                #                                   wavelet=wavelet,
                #                                   scales=[scale],
                #                                   sampling_period=1 / fs.value)
    # make unitary complex vector for phase estimation
    normalized_vector = y_filter / np.abs(y_filter)
    # if data comes in trials, then we average
    average_vector = np.mean(normalized_vector, axis=2)
    axs[0].plot(time, data, label='signal')
    axs[0].plot(time, np.real(reference_signal), label='ref')
    # compute phase difference
    # instantaneous_phase_difference = np.unwrap(np.angle(average_vector) * u.rad - np.angle(reference_signal), axis=0)
    instantaneous_phase_difference = (np.unwrap(np.angle(reference_signal), axis=0) -
                                      np.unwrap(np.angle(average_vector) * u.rad, axis=0)
                                      )
    axs[1].plot(time, instantaneous_phase_difference * 180 / np.pi, label='{:}'.format(phase * 180 / (np.pi * u.rad)))
    axs[1].legend()
plt.show()
plt.show()



import os
import numpy as np
import re
from peegy.definitions.channel_definitions import ChannelItem, ChannelType
import logging
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
log = logging.getLogger()


class DeviceEventChannel:
    bdf_event_channel = 'Status'
    edf_event_channel = 'EDF Annotations'
    bdf_event_annotations = 'BDF Annotations'
    auto = 'auto'


def read_edf_bdf_header(file_name):
    with open(file_name, 'rb') as f:
        # read first 256 bytes
        # first byte is skipped as is not important for bdf or edf files
        header = {'file_name': file_name,
                  'file_size': os.path.getsize(file_name),
                  'identification_code': int.from_bytes(f.read(1), byteorder='little'),
                  'version': f.read(7).strip().decode('ascii', errors='ignore'),
                  'subject_id': f.read(80).strip().decode('ascii', errors='ignore'),
                  'recording_id': f.read(80).strip().decode('ascii', errors='ignore'),
                  'start_date': f.read(8).strip().decode('ascii', errors='ignore'),
                  'start_time': f.read(8).strip().decode('ascii', errors='ignore'),
                  'bytes_in_header': int(f.read(8)),
                  'data_format': f.read(44).strip().decode('ascii', errors='ignore'),
                  'number_records': int(f.read(8)),
                  'duration_data_record': float(f.read(8)) * u.s,
                  'n_channels': int(f.read(4))}
        header['duration'] = header['number_records'] * header['duration_data_record']
        # read next N x 256 bytes
        ch = []
        [ch.append(ChannelItem(label=f.read(16).strip().decode('ascii', errors='ignore'), idx=i)) for i in
         range(header['n_channels'])]
        header['channels'] = np.array(ch)
        ch = []
        [ch.append(f.read(80).strip().decode('ascii', errors='ignore')) for i in range(header['n_channels'])]
        header['transducer'] = np.array(ch)
        ch = []
        [ch.append(f.read(8).strip().decode('ascii', errors='ignore')) for i in range(header['n_channels'])]
        header['physical_dimension'] = np.array(ch)
        ch = []
        [ch.append(float(f.read(8))) for i in range(header['n_channels'])]
        header['physical_minimum'] = np.array(ch)
        ch = []
        [ch.append(float(f.read(8))) for i in range(header['n_channels'])]
        header['physical_maximum'] = np.array(ch)
        ch = []
        [ch.append(float(f.read(8))) for i in range(header['n_channels'])]
        header['digital_minimum'] = np.array(ch)
        ch = []
        [ch.append(float(f.read(8))) for i in range(header['n_channels'])]
        header['digital_maximum'] = np.array(ch)
        ch = []
        [ch.append((header['physical_maximum'][i] - header['physical_minimum'][i]) /
                   (header['digital_maximum'][i] - header['digital_minimum'][i]))
         for i in range(header['n_channels'])]
        header['gain'] = np.array(ch)
        ch = []
        [ch.append(f.read(80).strip().decode('ascii', errors='ignore')) for i in range(header['n_channels'])]
        header['pre_filtering'] = np.array(ch)
        ch = []
        [ch.append(int(f.read(8))) for i in range(header['n_channels'])]
        header['number_samples_per_record'] = np.array(ch)
        ch = []
        [ch.append(f.read(32).strip().decode('ascii', errors='ignore')) for i in range(header['n_channels'])]
        header['reserved'] = np.array(ch)
        ch = []
        [ch.append((header['number_samples_per_record'][i] / header['duration_data_record']).to(u.Hz).value)
         for i in range(header['n_channels'])]
        header['fs'] = np.array(ch) * u.Hz
        if header['number_records'] == -1:
            header_off_set = 256 * (1 + header['n_channels'])
            header['number_records'] = ((os.stat(header['file_name']).st_size -
                                         header_off_set) / header['n_channels'] / header['fs'][0].value / 3
                                        ).astype(np.int64)
            header['duration'] = header['number_records'] * header['duration_data_record']

    return header


def read_bdf_channel(header,
                     channels_idx: np.array = np.array([]),
                     ini_time: u.quantity.Quantity = 0 * u.s,
                     end_time: u.quantity.Quantity = None,
                     include_event_channel: bool = True,
                     event_channel_label: DeviceEventChannel = DeviceEventChannel.bdf_event_channel,
                     data_unit: u.Unit = u.uV):
    fs_data = set_default_unit(header['fs'][0], u.Hz)
    ini_time = set_default_unit(ini_time, u.s)
    end_time = set_default_unit(end_time, u.s)
    if end_time is None:
        end_time = np.floor(header['duration'])
    else:
        end_time = np.minimum(end_time, header['duration'])

    print('Reading data in {:} units only'.format(data_unit))
    _idx_matching_units = np.array([])
    for _ch_idx, (_phy, _ch) in enumerate(zip(header['physical_dimension'], header['channels'])):
        if _ch.label != event_channel_label:
            try:
                _unit = u.Unit(_phy)
                if _unit.physical_type == data_unit.physical_type:
                    _idx_matching_units = np.append(_idx_matching_units, _ch_idx).astype(int)
            except ValueError:
                continue
        else:
            _idx_matching_units = np.append(_idx_matching_units, _ch_idx)

    if not channels_idx.size:
        channels_idx = np.arange(0, header['channels'].size).astype(int)

    channels_idx = np.intersect1d(channels_idx, _idx_matching_units)
    if not channels_idx.size:
        print('No data available in requested units {:}'.format(data_unit))
        return

    channels = header['channels'][channels_idx]
    annotated_event_channel = False
    fs_events = None
    if include_event_channel:
        if event_channel_label == DeviceEventChannel.bdf_event_annotations:
            annotated_event_channel = True

        # check if trigger channel is in channel list
        _idx_event = np.squeeze([_i for _i, _ch in enumerate(channels) if _ch.label == event_channel_label])
        if _idx_event.size == 0:
            _idx_event = [_i for _i, _ch in enumerate(header['channels']) if _ch.label == event_channel_label][0]
            channels = np.append(channels, header['channels'][_idx_event])
        fs_events = header['fs'][_idx_event]

    ini_time = np.round(ini_time / header['duration_data_record']) * header['duration_data_record']
    records_to_read = np.floor((end_time - ini_time) / header['duration_data_record']).astype(np.int64)
    # buffer_size = np.round(records_to_read * header['duration_data_record'] * fs_data).astype(np.int64)
    buffer_size_data = np.round(records_to_read * header['duration_data_record'] * fs_data).astype(np.int64)
    buffer_size_events = np.round(records_to_read * header['duration_data_record'] * fs_events).astype(np.int64)
    start_offset = (3 * ini_time.value * np.sum(header['number_samples_per_record'])).astype(np.int64)
    data_channel = np.zeros((buffer_size_data, len(channels) - 1), dtype=np.float32)
    event_channel = np.zeros((buffer_size_data, 1), dtype=np.float32)

    event_channel_annotations = None
    if annotated_event_channel:
        event_channel_annotations = np.zeros((buffer_size_events * 3, 1), dtype=np.uint8)
    # buffer to store events as additional channel
    sys_code_channel = np.zeros((buffer_size_data,), dtype=float)
    record_length = np.int64(np.sum(header['number_samples_per_record']) * 3)
    _time = ini_time
    with open(header['file_name'], 'rb') as f:
        # skip first 256 bytes + 256 * N channels
        header_off_set = 256 * (1 + header['n_channels'])
        ch_offsets = np.cumsum((np.concatenate(([0], header['number_samples_per_record']))))
        for nr in range(records_to_read):
            for i, ch in enumerate(channels):
                samples_per_record = np.int64(header['number_samples_per_record'][ch.idx])
                f.seek(header_off_set + start_offset + ch_offsets[ch.idx] * 3 + record_length * nr)
                data = np.fromfile(f, dtype=np.uint8, count=3 * samples_per_record)
                try:
                    if ch.label != event_channel_label or ch.label == DeviceEventChannel.bdf_event_channel:
                        data = data.reshape(-1, 3).astype(np.int32)
                except Exception:
                    print("Your data file seems to be INCOMPLETE, check it!")
                    break
                if not data.size:
                    break
                if ch.label != event_channel_label:
                    data = np.int32((data[:, 0] << 8) | (data[:, 1] << 16) | (data[:, 2] << 24)) >> 8
                    data_channel[nr * samples_per_record: (nr + 1) * samples_per_record, i] = data
                else:
                    if ch.label == DeviceEventChannel.bdf_event_channel:
                        sys_code_channel[nr * samples_per_record: (nr + 1) * samples_per_record] = data[:, 2]
                        # data = (data[:, 0] | (data[:, 1] << 8)) & 255
                        data = (np.int32((data[:, 0] << 8) | (data[:, 1] << 16) | (data[:, 2] << 24)) >> 8) & (
                                2 ** 16 - 1)
                        event_channel[nr * samples_per_record: (nr + 1) * samples_per_record, 0] = data
                        # set channel type to event
                        ch.type = ChannelType.Event
                    if ch.label == DeviceEventChannel.bdf_event_annotations:
                        event_channel_annotations[nr * samples_per_record * 3:
                                                  (nr + 1) * samples_per_record * 3, 0] = data
            _time = _time + header['duration_data_record']
            print('Processed time: {:.1f}'.format(_time))
    unique_events, unique_codes = None, None
    if annotated_event_channel:
        event_channel, unique_events, unique_codes = decode_annotated_events(
            data=event_channel_annotations,
            fs_data=fs_data,
            event_channel=event_channel,
            ini_time=ini_time
        )
    gain = np.array([header['gain'][ch.idx] for ch in channels])
    # convert all data channels to same unit (e.g. micro Volts)
    unit_scaling = np.array([u.Quantity(1, header['physical_dimension'][ch.idx]).to(data_unit).value
                             for ch in channels[0:-1]])
    scaled_data = data_channel * gain[0:-1] * unit_scaling * data_unit
    return (scaled_data,
            event_channel * gain[-1],
            (unique_events, unique_codes),
            channels_idx)


def read_edf_channel(header,
                     channels_idx: np.array = np.array([]),
                     ini_time: u.quantity.Quantity = 0 * u.s,
                     end_time: u.quantity.Quantity = None,
                     include_event_channel: bool = True,
                     event_channel_label: DeviceEventChannel = DeviceEventChannel.edf_event_channel,
                     data_unit: u.Unit = u.uV):
    fs_data = set_default_unit(header['fs'][0], u.Hz)
    ini_time = set_default_unit(ini_time, u.s)
    end_time = set_default_unit(end_time, u.s)
    if event_channel_label is None:
        event_channel_label = DeviceEventChannel.edf_event_channel
    if end_time is None:
        end_time = np.floor(header['duration'])
    else:
        end_time = np.minimum(end_time, header['duration'])

    print('Reading data in {:} units only'.format(data_unit))
    _idx_matching_units = np.array([])
    for _ch_idx, (_phy, _ch) in enumerate(zip(header['physical_dimension'], header['channels'])):
        if _ch.label != event_channel_label:
            try:
                _unit = u.Unit(_phy)
                if _unit.physical_type == data_unit.physical_type:
                    _idx_matching_units = np.append(_idx_matching_units, _ch_idx).astype(int)
            except ValueError:
                continue
        else:
            _idx_matching_units = np.append(_idx_matching_units, _ch_idx)

    if not channels_idx.size:
        channels_idx = np.arange(0, header['channels'].size).astype(int)

    channels_idx = np.intersect1d(channels_idx, _idx_matching_units)
    if not channels_idx.size:
        print('No data available in requested units {:}'.format(data_unit))
        return
    channels = header['channels'][channels_idx]

    _idx_event = None
    if include_event_channel:
        # check if trigger channel is in channel list
        _idx_event = np.squeeze(np.array([_i for _i, _ch in enumerate(channels) if _ch.label == event_channel_label]))
        if _idx_event.size == 0:
            _idx_event = [_i for _i, _ch in enumerate(header['channels']) if _ch.label == event_channel_label][0]
            channels = np.append(channels, header['channels'][_idx_event])
        fs_events = header['fs'][_idx_event]
    # ensure that in_time is multiple of duration_data_record
    ini_time = np.round(ini_time / header['duration_data_record']) * header['duration_data_record']
    records_to_read = np.floor((end_time - ini_time) / header['duration_data_record']).astype(np.int64)
    header_off_set = 256 * (1 + header['n_channels'])
    ch_offsets = np.cumsum((np.concatenate(([0], header['number_samples_per_record']))))

    buffer_size_data = np.round(records_to_read * header['duration_data_record'] * fs_data).astype(np.int64)
    buffer_size_events = np.round(records_to_read * header['duration_data_record'] * fs_events).astype(np.int64)
    data_channel = np.zeros((buffer_size_data, len(channels) - 1), dtype=np.float32)
    event_channel = np.zeros((buffer_size_data, 1), dtype=np.float32)
    sys_code_channel = np.zeros((buffer_size_events,), dtype=np.int16)
    record_length = np.sum(header['number_samples_per_record']) * 2
    start_offset = (ini_time.value * record_length).astype(np.int64)
    _time = ini_time
    unique_events, unique_codes = [], []
    with open(header['file_name'], 'rb') as f:
        # skip first 256 bytes + 256 * N channels
        for nr in range(records_to_read):
            for i, ch in enumerate(channels):
                samples_per_record = header['number_samples_per_record'][ch.idx]
                f.seek(header_off_set + start_offset + ch_offsets[ch.idx] * 2 + record_length * nr)
                data = np.fromfile(f, dtype=np.int16, count=samples_per_record)
                _length = data.shape[0]
                _ini_pos = nr * samples_per_record
                _end_pos = _ini_pos + _length
                if ch.label != (event_channel_label or DeviceEventChannel.edf_event_channel):
                    data_channel[_ini_pos: _end_pos, i] = data
                if ch.label == (event_channel_label or DeviceEventChannel.edf_event_channel):
                    sys_code_channel[_ini_pos: _end_pos] = data
                    # set channel type to event
                    ch.type = ChannelType.Event
            _time = _time + header['duration_data_record']
            print('Processed time {:}'.format(_time))

    if include_event_channel:
        if channels[-1].label == (event_channel_label or DeviceEventChannel.edf_event_channel):
            # fist we try getting annotated channel
            event_channel, unique_events, unique_codes = decode_annotated_events(
                data=sys_code_channel,
                fs_data=fs_data,
                event_channel=event_channel,
                ini_time=ini_time
            )
    gain = np.array([header['gain'][ch.idx] for ch in channels])
    # convert all data channels to same unit (e.g., micro Volts)
    unit_scaling = np.array([u.Quantity(1, header['physical_dimension'][ch.idx]).to(data_unit).value
                             for ch in channels[0:-1]])
    scaled_data = data_channel * unit_scaling * data_unit * gain[0:-1]
    return (scaled_data,
            event_channel,
            (unique_events, unique_codes),
            channels_idx)


def get_event_channel(header,
                      ini_time: u.quantity.Quantity = 0 * u.s,
                      end_time: u.quantity.Quantity = None,
                      event_channel_label=None):
    event_channel = None
    _, file_extension = os.path.splitext(header['file_name'])
    if event_channel_label is None:
        event_channel_label = DeviceEventChannel.bdf_event_channel if file_extension == '.bdf' \
            else DeviceEventChannel.edf_event_channel
    annotations = None
    _idx_event = np.array([_i for _i, _ch in enumerate(header['channels']) if _ch.label == event_channel_label])
    if file_extension == '.bdf':
        _, event_channel, annotations, _ = read_bdf_channel(header=header,
                                                            ini_time=ini_time,
                                                            end_time=end_time,
                                                            channels_idx=_idx_event,
                                                            event_channel_label=event_channel_label)
    if file_extension == '.edf':
        _, event_channel, annotations, _ = read_edf_channel(header=header,
                                                            ini_time=ini_time,
                                                            end_time=end_time,
                                                            channels_idx=_idx_event,
                                                            event_channel_label=event_channel_label)

    return event_channel, annotations


def get_data(header,
             channels_idx: np.array = None,
             ini_time: u.Quantity = 0 * u.s,
             end_time: u.Quantity = None,
             event_channel_label: DeviceEventChannel = None,
             data_unit: u.Unit = u.uV):
    data, events = None, None
    _, file_extension = os.path.splitext(header['file_name'])
    annotations = ()
    if file_extension == '.bdf':
        data, events, annotations, valid_idx = read_bdf_channel(header=header, channels_idx=channels_idx,
                                                                ini_time=ini_time, end_time=end_time,
                                                                event_channel_label=event_channel_label,
                                                                data_unit=data_unit)
    if file_extension == '.edf':
        data, events, annotations, valid_idx = read_edf_channel(header=header,
                                                                channels_idx=channels_idx,
                                                                ini_time=ini_time,
                                                                end_time=end_time,
                                                                event_channel_label=event_channel_label,
                                                                data_unit=data_unit)
    return data, events, data.unit, annotations, valid_idx


def decode_annotated_events(data: np.array = None,
                            fs_data: float = None,
                            ini_time: float = None,
                            event_channel: np.array = None):
    pattern = re.compile('([+-]\\d+\\.?\\d*)(\x15(\\d+\\.?\\d*))?(\x14.*?)\x14\x00')
    start_event = pattern.findall(data.tostring().decode('latin-1'))
    used_events = []
    for _ev in start_event:
        onset = float(_ev[0]) * u.s
        duration = float(_ev[2]) * u.s if _ev[2] else 0 * u.s
        for description in _ev[3].split('\x14')[1:]:
            if description:
                used_events.append([onset, duration, description])

    if len(used_events) > 0:
        event_pos = []
        event_dur = []
        event_label = []

        for _i_time, _dur, _code in used_events:
            event_pos.append(_i_time - ini_time)
            event_dur.append(0 if _dur == '' else _dur)
            event_label.append(_code.replace('\x14', ''))

        try:
            # if labels are integer, we use them as numeric code
            event_code = np.array(event_label).astype(int)
        except ValueError:
            # if labels are not, generate a numeric code
            unique_events = np.unique(event_label)
            unique_codes = np.arange(0, unique_events.size) + 1
            event_code = np.array([int(unique_codes[np.where(_l == unique_events)[0]]) for _l in event_label])

        for _ini, _dur, _code in zip(event_pos, event_dur, event_code):
            _ini_time = _ini
            _end_time = _ini_time + np.maximum(_dur, 1 / fs_data)
            # compute samples as this were recorded as a data channel
            _ini_sample = np.minimum((_ini_time * fs_data).astype(np.int64), event_channel.shape[0])
            _end_sample = np.minimum(np.maximum((_end_time * fs_data).astype(np.int64), _ini_sample + 1),
                                     event_channel.shape[0])

            event_channel[_ini_sample: _end_sample, 0] = _code
    else:
        # if none annotations are found we try to read channel  event as a data channel
        event_channel[:, 0] = data
    return event_channel, unique_events, unique_codes



import numpy as np
import math
from scipy.signal import firwin2
# from scipy.signal import freqz
from peegy.processing.tools.multiprocessing .multiprocessesing_filter import filt_data
import astropy.units as u


def rms(signal):
    return np.mean(signal ** 2.0, axis=0) ** 0.5


def lcm(num1, num2):
    return num1 * num2 / math.gcd(num1, num2)


def next_power_2(x: int):
    value = 1
    while value <= x:
        value = value << 1
    return value


def fit_spectrum(time_signal, time_shape_constrain, target_spectrum):
    n = time_signal.shape[0]

    if np.mod(n, 2) == 0:
        n_uniq_freq = int((n / 2 + 1))
    else:
        n_uniq_freq = int((n + 1) / 2)

    error = np.inf
    tol = 1e-10
    delta_error = np.inf
    _fft = np.fft.rfft(time_signal, axis=0)
    n_iter = 1
    fitted_signal = None
    while (delta_error > tol) and (n_iter < 1000):
        _phase = np.angle(_fft)
        # reconstruct full fft
        full_spectrum = target_spectrum * np.exp(1j * _phase)
        fitted_signal = np.fft.irfft(full_spectrum, n, axis=0) * time_shape_constrain
        _fft = np.fft.rfft(fitted_signal, axis=0)
        current_error = np.std(np.abs(_fft[range(n_uniq_freq)] - target_spectrum))
        delta_error = abs(error - current_error)
        error = current_error
        n_iter = n_iter + 1
        print('error: {:.6f}'.format(error))

    print('n_iter: {:}'.format(n_iter))
    return fitted_signal


def generate_modulated_noise(fs: float = 44100.0,
                             duration: float = 1.0,
                             n_channels=1,
                             n_repetitions: int = 1,
                             amplitude: float = 1.0,
                             f_noise_low: float = 300.0,
                             f_noise_high: float = 700.0,
                             attenuation: float = 0.0,
                             modulation_frequency: float = 0.0,
                             modulation_phase: float = 0.0,
                             modulation_index: float = 0.0,
                             reference_rms: bool = None,
                             noise_seed=None,
                             frequency_domain=False,
                             freq_resolution: float = 0.1,
                             n_taps: int = None,
                             offset_at_zero: int = 1):
    """
    This function generates coloured noise (1 / f ** p) in the frequency-domain.
    :param fs:  sampling rate
    :param duration: duration in seconds
    :param n_channels: number of output channels
    :param n_repetitions: how many times to repeat the time series (output is equal to duration * n_repetitions)
    :param amplitude: output maximum amplitude (between -1 and 1)
    :param f_noise_low: low frequency cutoff frequency
    :param f_noise_high: high frequency cutoff frequency
    :param attenuation: decay of power spectral density in dB / octave (3 dB for pink noise, 6 dB for brown noise)
    :param modulation_frequency: frequency in Hz to apply a temporal modulation to the noise
    :param modulation_phase: starting phase (in radians) of the modulation frequency
    :param modulation_index: value between 0 (un-modulated) and 1 (full modulated)
    :param reference_rms: rms value to apply in case of this value has been calculated to produced calibrated sounds
    :param noise_seed: float to initialize the random generator
    :param frequency_domain: if True, the noise is generated by randomizing the phase of the frequency components. If
    false, a FIR filter is estimated and applied to a random noise
    :param n_taps: integer indicating number of filter tabs in case of frequency_domain=False
    :param freq_resolution: frequency resolution to create FIR filter. If n_tabs is not passed, then number of tabs is
    computed to ensure the frequency resolution passed.
    :return: noise waveform (samples x channels)
    """
    if noise_seed is not None:
        np.random.seed(noise_seed)
    time = np.expand_dims(np.arange(0, np.round(fs * duration)) / fs, axis=1)
    n = time.size
    if np.mod(n, 2) == 0:
        n_uniq_freq = int((n / 2 + 1))
    else:
        n_uniq_freq = int((n + 1) / 2)

    freq = np.arange(0, n_uniq_freq).reshape([-1, 1]) * fs / n
    p = -attenuation / (10 * np.log10(0.5))
    psd_noise = np.zeros((n_uniq_freq, 1))

    # synthesize power spectral density
    _idx_power = np.argwhere(np.logical_and(freq >= f_noise_low, freq <= f_noise_high))[:, 0]
    psd_noise[_idx_power] = 1 / (offset_at_zero + freq[_idx_power] ** p)
    # ensure odd number of taps
    if n_taps is not None:
        n_taps = 2 * (n_taps // 2) + 1
    # synthesize phase
    if not frequency_domain:
        gain = np.sqrt(psd_noise)
        if freq_resolution is not None:
            n_taps = int(np.maximum(1, np.round(fs / freq_resolution)))
            n_taps = 2 * (n_taps // 2) + 1
        # force last frequency to be fs / 2
        freq[-1, :] = fs / 2
        print('noise generation: {:} taps FIR design'.format(n_taps))
        _b = firwin2(numtaps=n_taps,
                     freq=freq[:, 0],
                     fs=fs,
                     gain=gain[:, 0],
                     antisymmetric=False)
        white_noise = np.random.randn(n, n_channels)
        noise = filt_data(b=_b, data=white_noise * u.dimensionless_unscaled, mode='original').value
        # plot results
        # f, h = freqz(_b, worN=np.linspace(0, fs / 2, n_taps), fs=fs)
        # import matplotlib.pyplot as plt
        # plt.plot(f, 20 * np.log10(np.abs(h)), label='Filter response')
        # plt.plot(freq.flatten(), 20 * np.log10(gain), label='Target response')
    else:
        phase_noise = 2 * np.pi * np.random.rand(n_uniq_freq, n_channels)
        spectrum_noise = np.sqrt(psd_noise) * np.exp(1j * phase_noise)
        # synthesize noise in the time-domain
        noise = np.fft.irfft(spectrum_noise, n, axis=0)

    # modulated noise
    mod_amp = (1 - modulation_index * np.cos(2 * np.pi * modulation_frequency * time + modulation_phase)) / \
              (1 + modulation_index)
    _amplitude = amplitude * mod_amp
    noise = noise / np.max(np.abs(noise), axis=0)
    noise = _amplitude * noise

    # normalize noise to have required rms
    if reference_rms:
        noise = reference_rms / rms(noise) * noise
    value = np.tile(noise, n_repetitions)
    return value



import copy
import numpy as np
import datetime
from peegy.definitions.channel_definitions import Domain, ChannelItem
from peegy.io.eeg.reader import eeg_reader
from peegy.io.elipse_tools import parse_eclipse_data
from peegy.processing.pipe.definitions import InputOutputProcess, DataNode
from peegy.layouts import layouts
import astropy.units as u
from peegy.directories.tools import DirectoryPaths
from peegy.processing.events.event_tools import detect_events, get_event_times
import pandas as pd
from peegy.definitions.events import Events
from pathlib import Path
from os.path import sep
from peegy.tools.units.unit_tools import set_default_unit
from peegy.definitions.events import SingleEvent
import pyqtgraph as pg
pg.setConfigOption('leftButtonPan', False)


class ReadInputData(InputOutputProcess):
    def __init__(self,
                 file_path: str = None,
                 channels_idx: np.array = np.array([]),
                 ini_time: u.quantity.Quantity = 0 * u.s,
                 end_time: u.quantity.Quantity = np.Inf * u.s,
                 layout_file_name: str = None,
                 figures_folder: str = None,
                 figures_subset_folder: str = '',
                 fs_col_name: str = None,
                 gain_col_name: str = None,
                 fs_unit: u.quantity.Quantity = u.Hz,
                 gain_unit: u.quantity.Quantity = u.uV,
                 gain_inverted: bool = False,
                 event_channel_label: str = None,
                 data_unit: u.Unit = u.uV
                 ) -> InputOutputProcess:
        """
        This pipeline class handles reading eeg data files.

        :param file_path: path to file to be read
        :param channels_idx: numpy array indicating specifics channels to be read. If empty, all channels will be read
        :param ini_time: time in seconds from where to read data
        :param end_time: time in seconds up to where to read data
        :param layout_file_name: Name of layout mapping the channel labels to a specific topographic map
        :param figures_subset_folder: string used to generate a sub-folder within the main figure path. Useful to
        generate specific folder paths in a dynamic way.
        :param fs_col_name: provide the name of the column containing the sampling rate (used when data comes in .csv
        files)
        :param gain_col_name: provide the name of the column containing the gain to scale the data (used when data comes
         in .csv files)
        :param fs_unit: unit of sampling rate
        :param gain_unit: unit of gain
        :param gain_inverted: indicates whether the data will be multiplied or divided by the gain.
        :param event_channel_label: event_channel_label: string indicating the label of the channel containing all
         events. This is usually the 'Status' channel for bdf files and 'EDF Annotations' for EDF files, however, some
        devices provide other labels, e.g. 'BDF Annotations', therefore we leave this open
        :param data_unit: defines what data will be read from the files. This allows for filtering data with mixing
        units (uV, Celciusm, uS, etcetera).
        """
        super(ReadInputData, self).__init__()
        self.reader = None
        self.file_path = file_path
        self.channels_idx = channels_idx
        self.ini_time = ini_time
        self.end_time = end_time
        self.input_node = None
        self.layout_file_name = layout_file_name
        self.figures_subset_folder = figures_subset_folder
        self.figures_folder = figures_folder
        self.fs_col_name = fs_col_name
        self.gain_col_name = gain_col_name
        self.fs_unit = fs_unit
        self.gain_unit = gain_unit
        self.gain_inverted = gain_inverted
        self.event_channel_label = event_channel_label
        self.data_unit = data_unit

    def transform_data(self):
        self.reader = eeg_reader(file_name=self.file_path,
                                 fs_col_name=self.fs_col_name,
                                 gain_col_name=self.gain_col_name,
                                 fs_unit=self.fs_unit,
                                 gain_unit=self.gain_unit,
                                 gain_inverted=self.gain_inverted,
                                 event_channel_label=self.event_channel_label,
                                 data_unit=self.data_unit)
        data, events, units, annotations = self.reader.get_data(channels_idx=self.channels_idx,
                                                                ini_time=self.ini_time,
                                                                end_time=self.end_time)
        self.output_node = DataNode(data=data,
                                    fs=self.reader.fs,
                                    domain=Domain.time,
                                    layout=self.reader.default_layout,
                                    paths=DirectoryPaths(file_path=self.reader.file_name,
                                                         delete_all=False,
                                                         delete_figures=False,
                                                         figures_folder=self.figures_folder,
                                                         figures_subset_folder=self.figures_subset_folder)
                                    )
        if self.layout_file_name is not None:
            layout = layouts.Layout(file_name=self.layout_file_name)
            self.output_node.apply_layout(layout)
        self.output_node.events = self.get_events(events)
        self.output_node.events_annotations = annotations
        if annotations[0] is not None and annotations[1] is not None:
            to_print = pd.DataFrame.from_dict(
                {'Annotation': annotations[0],
                 'Code': annotations[1]})
            print(to_print.to_string())

    def get_events(self, events):
        _events = get_event_times(event_channel=events, fs=self.output_node.fs)
        return _events


class MergeMultipleFiles(InputOutputProcess):
    def __init__(self,
                 file_paths: [str] = None,
                 channels_idx: np.array = np.array([]),
                 layout_file_name: str = None,
                 figures_subset_folder: str = '',
                 fs_col_name: str = None,
                 gain_col_name: str = None,
                 fs_unit: u.quantity.Quantity = u.Hz,
                 gain_unit: u.quantity.Quantity = u.uV,
                 gain_inverted: bool = False
                 ) -> InputOutputProcess:
        super(MergeMultipleFiles, self).__init__()
        self.readers = np.array([])
        for _file in file_paths:
            _reader = eeg_reader(_file,
                                 fs_col_name=fs_col_name,
                                 gain_col_name=gain_col_name,
                                 fs_unit=fs_unit,
                                 gain_unit=gain_unit,
                                 gain_inverted=gain_inverted)
            self.readers = np.append(self.readers, _reader)
        self.file_paths = file_paths
        self.channels_idx = channels_idx
        self.ini_time = 0 * u.s
        self.end_time = np.Inf * u.s
        self.output_node = None
        self.layout_file_name = layout_file_name
        self.figures_subset_folder = figures_subset_folder
        self.output_node = DataNode(fs=self.readers[0].fs,
                                    domain=Domain.time,
                                    layout=self.readers[0].default_layout,
                                    paths=DirectoryPaths(file_path=self.readers[0].file_name,
                                                         delete_all=False, delete_figures=False,
                                                         figures_subset_folder=figures_subset_folder)
                                    )

    def transform_data(self):
        all_data = None
        all_events = None
        # sort readers by time
        _date_format = "%d.%m.%y/%H.%M.%S"
        dates = np.array([_reader._header['start_date'] + '/' + _reader._header['start_time'] for
                          _reader in self.readers])
        sorted_idx = np.argsort([datetime.datetime.strptime(_date, _date_format) for _date in dates])
        print(dates[sorted_idx])
        for _reader in self.readers[sorted_idx]:
            data, events, units, annotations = _reader.get_data(channels_idx=self.channels_idx,
                                                                ini_time=self.ini_time,
                                                                end_time=self.end_time)

            # demean data
            data = data - np.mean(data, axis=0)
            if all_data is not None:
                all_data = np.concatenate((all_data, data))
                all_events = np.concatenate((all_events, events))
            else:
                all_data = data
                all_events = events

        self.output_node = DataNode(data=all_data,
                                    fs=self.readers[0].fs,
                                    domain=Domain.time,
                                    layout=self.readers[0].default_layout,
                                    )
        if self.layout_file_name is not None:
            self.output_node.apply_layout(layouts.Layout(file_name=self.layout_file_name))

        self.output_node.paths = self.input_node.paths
        self.get_events(all_events)

    def get_events(self, events):
        events = detect_events(event_channel=events, fs=self.output_node.fs)
        events = Events(events=np.array(events))
        for i, _code in enumerate(np.unique(events.get_events_code())):
            print('Event code:', _code, 'Number of events:', events.get_events_code(code=_code).size)
        self.output_node.events = events


class GenericInputData(InputOutputProcess):
    def __init__(self,
                 data: np.array = None,
                 fs: u.quantity.Quantity = 16384.0 * u.Hz,
                 events_index: int = None,
                 event_times: np.array = None,
                 event_code: float = 1.0,
                 figures_path: str = None,
                 channel_labels: [str] = None,
                 figures_subset_folder: str = '') -> InputOutputProcess:
        """
        This class allows to pass your own data without the need of having a bdf or edf file. Data will be used to
        create a compatible InoutOutputProcess that can be use straightforward in the pipeline.
        This InputOutput process takes a numpy matrix and uses it to generate a generic layout.
        :param data: numpy array (2D or 3D array; samples x channels x trials)
        :param fs: the sampling rate of the template_waveform
        :param events_index: integer in which column are the events included
        :param event_times: numpy array with the timing of the events. Events are only useful when input data is a 2D
        numpy array.
        :param event_code: desired event code to be assigned to time events
        :param figures_path: path to save generated figures
        :param figures_subset_folder: string indicating a sub-folder name in figures_path
        """
        super(GenericInputData, self).__init__()
        events = None
        if events_index is not None:
            events = data[:, np.arange(data.shape[1]) == events_index]
            data = data[:, np.arange(data.shape[1]) != events_index]
        self.events = events
        self.data = set_default_unit(copy.copy(data), u.uV)
        self.fs = set_default_unit(fs, u.Hz)
        self.event_times = set_default_unit(event_times, u.s)
        self.event_code = event_code
        self.output_node = None
        self.figures_path = figures_path if figures_path is not None else str(Path.home()) + '{:}'.format(sep +
                                                                                                          'peegy' +
                                                                                                          sep +
                                                                                                          'test' +
                                                                                                          sep +
                                                                                                          'figures')
        self.figures_subset_folder = figures_subset_folder

        _ch = []
        n_channels = self.data.shape[1]
        if channel_labels is None:
            [_ch.append(ChannelItem(label='CH_{:}'.format(i), idx=i)) for i in range(n_channels)]
        else:
            [_ch.append(ChannelItem(label=_label, idx=i)) for i, _label in
             zip(range(n_channels), channel_labels)]
        layout = np.array(_ch)
        self.output_node = DataNode(fs=fs,
                                    domain=Domain.time,
                                    layout=layout,
                                    paths=DirectoryPaths(file_path=self.figures_path,
                                                         delete_all=False,
                                                         delete_figures=False,
                                                         figures_subset_folder=figures_subset_folder),
                                    )

    def transform_data(self):
        events = np.array([])
        if self.event_times is not None:
            for _ev in self.event_times:
                events = np.append(events, SingleEvent(code=self.event_code,
                                                       time_pos=_ev,
                                                       dur=0))
        events = Events(events=np.array(events))
        if self.events is not None:
            events = self.get_events(self.events)
        self.output_node = DataNode(data=self.data,
                                    fs=self.fs,
                                    domain=Domain.time,
                                    layout=self.input_node.layout,
                                    )
        self.output_node.paths = self.input_node.paths
        self.output_node.events = events

    def get_events(self, events):
        _events = get_event_times(event_channel=events, fs=self.fs)
        return _events


class EclipseReader(InputOutputProcess):
    def __init__(self,
                 file_path: np.array = None,
                 buffer: str = 'A_&_B',
                 figures_subset_folder: str = '') -> InputOutputProcess:
        """
        This class will parse Eclipse data contained in buffers A or B (provided in .tex files by Eclipse software)
        :param data_path: path to the file
        :param buffer: 'A', 'B', 'A_&_B' (or None). This will return the desired buffers contained in the data
        :param figures_path: path to save generated figures
        :param figures_subset_folder: string indicating a sub-folder name in figures_path
        """
        super(EclipseReader, self).__init__()
        self.header = None
        self.output_node = None
        self.buffer = buffer
        self.file_path = file_path
        self.figures_subset_folder = figures_subset_folder
        self.output_node = DataNode(fs=None,
                                    domain=Domain.time,
                                    layout=None,
                                    paths=DirectoryPaths(file_path=str(file_path),
                                                         delete_all=False,
                                                         delete_figures=False,
                                                         figures_subset_folder=figures_subset_folder)
                                    )

    def transform_data(self):
        events = np.array([])
        events = Events(events=np.array(events))
        header, data, data_buffer_a, data_buffer_b = parse_eclipse_data(
            file_name=self.file_path)
        if self.buffer == 'A':
            _data = data_buffer_a
            _ch = [ChannelItem(label='Buffer_A')]
            layout = np.array(_ch)
        elif self.buffer == 'B':
            _data = data_buffer_b
            _ch = [ChannelItem(label='Buffer_B')]
            layout = np.array(_ch)
        elif self.buffer == 'A_&_B' or None:
            _data = data
            _ch = [ChannelItem(label='Buffers_A_and_B')]
            layout = np.array(_ch)
            self.buffer == 'A_&_B'

        fs = header['Sample rate'] * u.Hz
        self.header = header
        self.output_node = DataNode(data=_data,
                                    fs=fs,
                                    domain=Domain.time,
                                    layout=layout,
                                    )
        self.output_node.paths = self.input_node.paths
        self.output_node.events = events
        self.output_node.x_offset = -header['Time offset in samples'] / fs



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools import epochs_processing_tools as ept
from peegy.processing.statistics import eeg_statistic_tools as espt
import numpy as np
import matplotlib.pyplot as plt
# create synthetic data
fs = 1000.0
nsamples = np.round(1 * fs).astype(int)
nchans = 36
ntrials = 130
noise_dim = 36  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs), axis=1)
coeff = np.ones(nchans//2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, coeff)), 0)

s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))

desired_snr = 15.0
ini_std = 10.0 ** (-desired_snr / 20.0) * s_std * ntrials ** 0.5
theoretical_rn = ini_std / ntrials ** 0.5

noise = np.random.normal(0, ini_std[0], size=(nsamples, nchans, ntrials))
# s[:, 0] = s[:, 0] * 0.5
data = noise + s

# get samples
block_size = int(0.040*fs)
samples = np.array([np.mean(data[_samp*block_size:(_samp + 1) * block_size, :, :], 0) for _samp in range(9)])
a = espt.hotelling_t_square_test(samples)
print(np.array([_x.p_value for _x in a]).T < 0.05)

roi_windows = [np.array([100, 200]), np.array([400, 500])]
w_ave, w, final_rn, cumulative_rn, w_fft, n, *_ = ept.et_mean(epochs=data,
                                                              block_size=10,
                                                              samples_distance=10
                                                              )

snr, _ = ept.et_snr_in_rois(w_ave,
                            roi_windows=roi_windows,
                            rn=final_rn)
print(["Theoretic RN", theoretical_rn, "Estimated rn", final_rn])
print(["Theoretic SRN", desired_snr, "Estimated srn", 10 * np.log10(snr)])

fig = plt.figure()
ax = fig.add_subplot(131)
ax.plot(np.mean(s, axis=2))
ax = fig.add_subplot(132)
ax.plot(w_ave)
ax = fig.add_subplot(133)
plt.plot(np.array(cumulative_rn))
plt.show()
plt.show()



import numpy as np
from peegy.definitions.channel_definitions import Domain, ChannelItem
from peegy.processing.pipe.definitions import InputOutputProcess, DataNode
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
from peegy.tools.signal_generator import noise_functions as nf
from peegy.layouts import layouts
import astropy.units as u
from peegy.directories.tools import DirectoryPaths
from peegy.processing.events.event_tools import detect_events
from peegy.definitions.events import Events
from pathlib import Path
from os.path import sep
from peegy.processing.tools.template_generator.auditory_waveforms import eog_template, fade_in_out_window
from peegy.tools.units.unit_tools import set_default_unit
from scipy.linalg import eigh, cholesky
import pyqtgraph as pg

pg.setConfigOption('leftButtonPan', False)


def correlated_noise(method: str = 'cholesky',
                     noise_token: np.array = None,
                     n_channels: int = 4,
                     covariance_matrix: np.array = None,
                     ):
    assert covariance_matrix.shape == (n_channels, n_channels)
    # We need a matrix `c` for which `c*c^T = r`.  We can use, for example,
    # the Cholesky decomposition, or we can construct `c` from the
    # eigenvectors and eigenvalues.
    # check if covariance matrix is unitary
    if np.all(covariance_matrix == 1):
        c = np.ones(covariance_matrix.shape)
    elif method == 'cholesky':
        # Compute the Cholesky decomposition.
        c = cholesky(covariance_matrix, lower=True)
    else:
        # Compute the eigenvalues and eigenvectors.
        evals, evecs = eigh(covariance_matrix)
        # Construct c, so c*c^T = r.
        c = np.dot(evecs, np.diag(np.sqrt(evals)))

    # Convert the data to correlated random variables.
    y = np.dot(c, noise_token.T).T
    return y


class GenerateInputData(InputOutputProcess):
    def __init__(self,
                 template_waveform: np.array = None,
                 stimulus_waveform: np.array = None,
                 fs: u.quantity.Quantity = 16384.0 * u.Hz,
                 alternating_polarity=False,
                 n_channels: int = None,
                 snr: float = 3,
                 line_noise_amplitude: u.quantity.Quantity = 0 * u.uV,
                 line_noise_mixing_matrix: np.array = None,
                 line_noise_frequency: u.Quantity = 50 * u.Hz,
                 line_noise_n_harmonics: int = 1,
                 include_non_stationary_noise_events=False,
                 noise_events_interval: u.quantity.Quantity = 25.0 * u.s,
                 noise_events_duration: u.quantity.Quantity = 15.0 * u.s,
                 noise_events_power_delta_db: float = 6.0,
                 noise_seed: float = None,
                 noise_attenuation: float = 0,
                 noise_covariance_matrix: np.array = None,
                 mixing_matrix: np.array = None,
                 artefacts_mixing_matrix: np.array = None,
                 eog_mixing_matrix: np.array = None,
                 return_eog_channel: bool = True,
                 include_eog_events: bool = False,
                 fade_in_out: bool = False,
                 event_times: np.array = None,
                 event_code: float = 1.0,
                 events_asymmetry_interval: int = 1,
                 events_asymmetry_factor: float = 1.0,
                 layout_file_name: str = None,
                 figures_path: str = None,
                 figures_subset_folder: str = '',
                 return_noise_only=False,
                 f_noise_low: u.Quantity = 0 * u.Hz,
                 f_noise_high: u.Quantity = None,
                 f_var_low: u.Quantity = 0 * u.Hz,
                 f_var_high: u.Quantity = None,
                 neural_ini_time_snr: u.Quantity = 0 * u.s,
                 neural_end_time_snr: u.Quantity = None,
                 noise_generation_length_factor: float = 1.0) -> InputOutputProcess:
        """
        This class allows to generate data that can be used without the need of having a bdf or edf file.
        This InputOutput process takes an input template and convolve it with delta at event times.
        The output will have as many channels as desired and the maximum snr will be given by snr.
        If requested, non-stationary noise events will be generated. The snr will be calculated relative to the
        stationary noise. The non-stationary events will vary around that stationary level by x dB as passed in
        noise_events_power_delta_db.
        The output events will be coded with the provided event code.
        :param template_waveform: a numpy array with the template waveform to be use as signal
        :param stimulus_waveform: a numpy array with the presented waveform. Useful to generate stimulus artifacts
        :param fs: the sampling rate of the template_waveform
        :param alternating_polarity: if true, stimulus_waveform is alternating in polarity for every epoch
        :param n_channels: number of channels to generate
        :param snr: signal-to-noise ratio defined as variance_signal / variance_noise, 0 < snr < inf.
        This will be used to set the amplitude of the stationary noise
        :param line_noise_amplitude: amplitude in uV of 50 Hz line noise
        :param line_noise_frequency: frequency of line noise
        :param line_noise_n_harmonics: number of harmonics to include in line noise
        :param include_non_stationary_noise_events: if True, non-stationary noise events will be included
        :param noise_events_interval: interval time [in seconds] between non-stationary events
        :param noise_events_duration: maximum duration of each non-stationary noise event
        :param noise_events_power_delta_db: the dB change of power in the noise noise relative to stationary noise power
        :param noise_seed: float to set the random generator
        :param noise_attenuation: indicates the attenuation of the noise (1 / f ** noise_attenuation), set to 0 for
        white noise and to 3 for pink noise
        :param noise_covariance_matrix: n_channels by n_channels matrix indicating the desired covariance of the noise
        :param mixing_matrix: numpy array with the coefficients that will be applied to template wavefrom
        :param include_eog_events: if True, eog-like artifacts will be included
        :param return_eog_channel: it True, eog waveforms will be appended as another channel
        :param fade_in_out: It True, convolved template will be faded in and out to prevent onset effects
        :param event_times: numpy array with the location of the events. The wave
        :param event_code: desired event code to be assigned to time events
        :param events_asymmetry_interval: integer indicating the number rate at which events delta pulses will be
        modulated with a value different from 1. This is useful to simulate an asymmetric neural response every
        events_asymmetry_interval events.
        :param events_asymmetry_factor: float: indicate the scaling factor of  the dirac pulse applied every
        events_asymmetry_interval.
        :param layout_file_name: name of layout to be used. e.g. "biosemi32.lay"
        :param figures_path: path to save generated figures
        :param figures_subset_folder: string indicating a subfolder name in figures_path
        :param return_noise_only: boolean indicating if the output will consist of noise only or not. This is useful to
        :param f_noise_low: low cuttoff frequency of generated noise
        :param f_noise_high: high cuttof frequency of generated noise
        :param f_var_low: low cuttoff to compute the energy of the noise. This is useful to produce noise where the
        energy needs to be equated in a frequency range.
        :param f_var_high: high cuttoff to compute the energy of the noise. This is useful to produce noise where the
        energy needs to be equated in a frequency range.
        :param noise_generation_length_factor is used to generate a noise that last longer than desired. This has the
        effect of increasing the density of the frequency components.
        compare results with and without a target signal whilst keeping the noise output fixed.
        """
        super(GenerateInputData, self).__init__()
        self.template_waveform = template_waveform
        self.stimulus_waveform = stimulus_waveform
        self.alternating_polarity = alternating_polarity
        self.n_channels = n_channels
        self.snr = snr
        self.line_noise_amplitude = set_default_unit(line_noise_amplitude, u.V)
        self.line_noise_frequency = set_default_unit(line_noise_frequency, u.Hz)
        self.line_noise_n_harmonics = line_noise_n_harmonics
        self.line_noise_mixing_matrix = line_noise_mixing_matrix
        self.fs = set_default_unit(fs, u.Hz)
        self.include_non_stationary_noise = include_non_stationary_noise_events
        self.noise_events_interval = set_default_unit(noise_events_interval, u.s)
        self.noise_events_duration = set_default_unit(noise_events_duration, u.s)
        self.noise_events_power_delta_db = noise_events_power_delta_db
        self.noise_seed = noise_seed
        self.noise_attenuation = noise_attenuation
        self.noise_covariance_matrix = set_default_unit(noise_covariance_matrix, template_waveform.unit ** 2.0)
        self.noise_generation_length_factor = noise_generation_length_factor
        self.event_times = set_default_unit(event_times, u.s)
        self.event_code = event_code
        self.events_asymmetry_interval = events_asymmetry_interval
        self.events_asymmetry_factor = events_asymmetry_factor
        self.include_eog_events = include_eog_events
        self.simulated_artifact = None
        self.simulated_neural_response = None
        self.simulated_artifact_response = None
        self.output_node = None
        self.layout_file_name = layout_file_name
        self.mixing_matrix = mixing_matrix
        self.artefacts_mixing_matrix = artefacts_mixing_matrix
        self.eog_mixing_matrix = eog_mixing_matrix
        self.return_eog_channel = return_eog_channel
        self.fade_in_out = fade_in_out
        self.return_noise_only = return_noise_only
        self.f_noise_low = set_default_unit(f_noise_low, u.Hz)
        self.f_noise_high = set_default_unit(f_noise_high, u.Hz)
        self.f_var_low = set_default_unit(f_var_low, u.Hz)
        self.f_var_high = set_default_unit(f_var_high, u.Hz)
        self.neural_ini_time_snr = neural_ini_time_snr
        self.neural_end_time_snr = neural_end_time_snr
        figures_path = figures_path if figures_path is not None else str(Path.home()) + '{:}'.format(
            sep +
            'peegy' +
            sep +
            'test' +
            sep +
            'figures')
        self.figures_subset_folder = figures_subset_folder

        _ch = []
        [_ch.append(ChannelItem(label='CH_{:}'.format(i), idx=i)) for i in range(n_channels)]
        layout = np.array(_ch)
        self.output_node = DataNode(fs=fs,
                                    domain=Domain.time,
                                    layout=layout,
                                    paths=DirectoryPaths(file_path=figures_path,
                                                         delete_all=False,
                                                         delete_figures=False,
                                                         figures_subset_folder=figures_subset_folder)
                                    )

    def transform_data(self):
        if self.noise_seed is not None:
            np.random.seed(self.noise_seed)
        # create synthetic data
        _events_idx = np.floor(self.event_times * self.fs).astype(int)
        events = np.zeros((int(self.event_times[-1] * self.fs) + 1, 1))
        events[_events_idx, :] = 1
        events[_events_idx[0::self.events_asymmetry_interval], :] = self.events_asymmetry_factor
        source = self.template_waveform
        # convolve source with dirac train to generate entire signal
        source = filt_data(data=source, b=events.flatten(), mode='full', onset_padding=False)
        fade_window = np.ones(source.shape)
        if self.fade_in_out:
            fade_window = fade_in_out_window(n_samples=source.shape[0],
                                             fraction=10 * self.template_waveform.shape[0] / source.shape[0])
        source = source * fade_window
        # generate the source in all channels
        if self.mixing_matrix is None:
            self.mixing_matrix = np.diag(np.ones(self.n_channels))
        tailed_source = source
        if tailed_source.shape[1] != self.n_channels:
            tailed_source = np.tile(source, (1, self.n_channels))
        assert tailed_source.shape[1] == self.mixing_matrix.shape[1]
        neural_response = tailed_source.dot(self.mixing_matrix)
        self.simulated_neural_response = neural_response

        # artifact mixing matrix
        mixed_stimulus_artefacts = np.zeros(neural_response.shape)
        artefact_source = None
        if self.stimulus_waveform is not None:
            artefact_source, mixed_stimulus_artefacts = generate_arbitrary_artefacts(
                stimulus_waveform=self.stimulus_waveform,
                n_channels=self.n_channels,
                events=events.copy(),
                alternating_polarity=self.alternating_polarity,
                artefacts_mixing_matrix=self.artefacts_mixing_matrix)
            self.simulated_artifact = artefact_source
            self.simulated_artifact_response = mixed_stimulus_artefacts

        # eog mixing matrix
        eog_mixed_artefacts = np.zeros(source.shape) * u.uV
        if self.include_eog_events:
            eog_waveform, eog_mixed_artefacts = generate_eog_artefacts(size=source.shape[0],
                                                                       n_channels=self.n_channels,
                                                                       eog_mixing_matrix=self.eog_mixing_matrix,
                                                                       fs=self.fs)

        # generate eeg noise
        noise, noise_var = generate_eeg_noise(f_noise_high=self.f_noise_high,
                                              f_noise_low=self.f_noise_low,
                                              f_var_high=self.f_var_high,
                                              f_var_low=self.f_var_low,
                                              n_channels=self.n_channels,
                                              fs=self.fs,
                                              noise_attenuation=self.noise_attenuation,
                                              noise_seed=self.noise_seed,
                                              noise_covariance_matrix=self.noise_covariance_matrix,
                                              size=source.shape[0],
                                              noise_generation_length_factor=self.noise_generation_length_factor,
                                              include_non_stationary_noise=self.include_non_stationary_noise,
                                              noise_events_duration=self.noise_events_duration,
                                              noise_events_interval=self.noise_events_interval,
                                              noise_events_power_delta_db=self.noise_events_power_delta_db
                                              )

        # generate line noise
        if self.line_noise_mixing_matrix is None:
            self.line_noise_mixing_matrix = np.diag((1 + np.random.rand(self.n_channels) * 0.001))
        line_artefact = generate_line_noise(size=source.shape[0],
                                            fs=self.fs,
                                            n_channels=self.n_channels,
                                            line_noise_amplitude=self.line_noise_amplitude,
                                            line_noise_frequency=self.line_noise_frequency,
                                            line_noise_n_harmonics=self.line_noise_n_harmonics,
                                            line_noise_mixing_matrix=self.line_noise_mixing_matrix)

        # compute variance of signal
        _ini_sample = int(self.neural_ini_time_snr * self.fs)
        _end_sample = neural_response.shape[0]
        if self.neural_end_time_snr is not None:
            _end_sample = np.minimum(int(self.neural_end_time_snr * self.fs), neural_response.shape[0])
        neural_var = np.var(neural_response[_ini_sample: _end_sample], ddof=1, axis=0)
        # scale signal

        scaled_noise = (neural_var / (self.snr * noise_var)) ** 0.5 * noise
        scaled_noise_variance = np.var(scaled_noise, ddof=1, axis=0) + np.finfo(float).eps * scaled_noise.unit ** 2.0
        data = (neural_response * (not self.return_noise_only) +
                scaled_noise +
                line_artefact +
                eog_mixed_artefacts +
                mixed_stimulus_artefacts)

        print('Theoretical RN for {:} events is {:}, and SNR {:} [dB] when target signal is included'.format(
            self.event_times.shape[0],
            np.sqrt(scaled_noise_variance / self.event_times.shape[0]),
            10 * np.log10(self.event_times.shape[0] * neural_var / scaled_noise_variance + np.finfo(float).eps)))
        events[_events_idx.astype(int)] = self.event_code

        self.output_node = DataNode(data=data,
                                    fs=self.fs,
                                    domain=Domain.time,
                                    layout=self.input_node.layout,
                                    )
        if self.layout_file_name is not None:
            self.output_node.apply_layout(layouts.Layout(file_name=self.layout_file_name))

        self.output_node.paths = self.input_node.paths
        self.get_events(events)
        if self.include_eog_events and self.return_eog_channel:
            self.output_node.append_new_channel(new_data=eog_waveform,
                                                layout_label='EOG1')

    def get_events(self, events):
        events = detect_events(event_channel=events, fs=self.output_node.fs)
        events = Events(events=np.array(events))
        for i, _code in enumerate(np.unique(events.get_events_code())):
            print('Event code:', _code, 'Number of events:', events.get_events_code(code=_code).size)
        self.output_node.events = events


def generate_eog_artefacts(size: int = None,
                           n_channels: int = None,
                           eog_mixing_matrix: np.array = None,
                           fs: u.Quantity = None,
                           events_mean_time: u.Quantity = 2 * u.s,
                           events_std_time: u.Quantity = 1.4 * u.s):
    if eog_mixing_matrix is None:
        eog_mixing_matrix = np.diag((np.random.rand(n_channels) - 0.5) / 0.5)
    # generates and convolve eye artifacts
    oeg_events = np.zeros((size, 1))
    eog_time_events = np.cumsum(
        events_std_time * np.random.randn(int((size/fs) / events_mean_time)) + events_mean_time)
    eog_time_events = np.minimum(np.maximum(eog_time_events, 0), np.floor((size - 1) / fs))
    _oeg_events_idx = np.floor(eog_time_events * fs).astype(int)
    oeg_events[_oeg_events_idx, :] = 1
    # _oeg_template, _ = eog(fs)
    _oeg_template, _ = eog_template(fs)
    oeg_events = filt_data(data=_oeg_template,
                           b=oeg_events.flatten(),
                           mode='full',
                           onset_padding=False)
    eog_waveform = oeg_events[0: size, :]
    eog_mixed_artefacts = np.tile(eog_waveform, (1, n_channels)).dot(eog_mixing_matrix)
    return eog_waveform, eog_mixed_artefacts


def generate_arbitrary_artefacts(stimulus_waveform: np.array = None,
                                 n_channels: int = None,
                                 events: np.array = None,
                                 alternating_polarity: bool = False,
                                 artefacts_mixing_matrix: np.array = None):
    if artefacts_mixing_matrix is None:
        artefacts_mixing_matrix = np.diag(np.ones(n_channels))
    events_idx = np.argwhere(events == 1)
    if alternating_polarity:
        events[events_idx[1:-1:2], :] = -1

    artefact_source = filt_data(data=stimulus_waveform,
                                b=events.flatten(),
                                mode='full',
                                onset_padding=False)
    tiled_artefact_source = artefact_source
    if tiled_artefact_source.shape[1] != n_channels:
        tiled_artefact_source = np.tile(artefact_source, (1, n_channels))
    assert tiled_artefact_source.shape[1] == n_channels
    mixed_artefacts = tiled_artefact_source.dot(artefacts_mixing_matrix)
    return artefact_source, mixed_artefacts


def generate_eeg_noise(f_noise_high: u.Quantity = None,
                       f_noise_low: u.Quantity = None,
                       f_var_high: u.Quantity = None,
                       f_var_low: u.Quantity = None,
                       fs: u.Quantity = None,
                       noise_attenuation: float = 0,
                       noise_seed: float = None,
                       noise_covariance_matrix: np.array = None,
                       n_channels: int = None,
                       size: int = None,
                       noise_generation_length_factor: float = 3,
                       include_non_stationary_noise: bool = False,
                       noise_events_interval: u.Quantity = None,
                       noise_events_duration: u.Quantity = None,
                       noise_events_power_delta_db: float = None,
                       ):
    if f_noise_high is None:
        f_noise_high = fs / 2
    else:
        f_noise_high = np.minimum(f_noise_high.to(u.Hz).value, fs.to(u.Hz).value / 2) * u.Hz
    if f_var_high is None:
        f_var_high = fs / 2
    else:
        f_var_high = np.minimum(f_var_high.to(u.Hz).value, fs.to(u.Hz).value / 2) * u.Hz

    # to do: use FIR filter to generate 1 /f ** p noise!
    noise = nf.generate_modulated_noise(
        fs=fs.to(u.Hz).value,
        f_noise_low=f_noise_low.to(u.Hz).value,
        f_noise_high=f_noise_high.to(u.Hz).value,
        duration=size / fs.to(u.Hz).value * noise_generation_length_factor,
        n_channels=n_channels,
        attenuation=noise_attenuation,
        noise_seed=noise_seed)
    # if noise_seed is not None:
    #     np.random.seed(noise_seed)
    # noise = np.random.normal(1, size=(size, n_channels))

    if noise_covariance_matrix is None:
        noise_covariance_matrix = np.diag(np.ones(n_channels))
    noise_covariance_matrix = set_default_unit(noise_covariance_matrix, u.dimensionless_unscaled)
    noise = noise[0: size, ...]
    noise = correlated_noise(noise_token=noise,
                             n_channels=n_channels,
                             covariance_matrix=noise_covariance_matrix.value)

    print('noise correlation: {:}'.format(np.corrcoef(noise.T)))

    noise = noise - np.mean(noise, axis=0)
    # normalize noise to have required rms
    freq = np.fft.fftfreq(noise.shape[0], 1) * fs
    noise_psd = 1 / (noise.shape[0] ** 2) * np.abs(np.fft.fft(noise, axis=0)) ** 2.0
    noise_psd[0, :] = noise_psd[0, :] / 2
    _idx_psd = np.argwhere(np.logical_and(freq >= f_var_low, freq <= f_var_high))[:, 0]
    stationary_noise_var = np.sum(2 * noise_psd[_idx_psd, :], axis=0)

    # stationary_noise_var = np.var(noise, ddof=0, axis=0)
    if include_non_stationary_noise:
        # non-stationary noise (e.g. 1 event every 20 seconds)
        _new_noise_events = np.arange(0, size,
                                      noise_events_interval * fs) / fs
        for _i, _ne in enumerate(_new_noise_events):
            _ini_pos = int(_ne * fs)
            _noise_samples = np.random.randint(1, int(fs * noise_events_duration))
            _ini_pos = np.minimum(_ini_pos, size)
            _end_pos = np.minimum(_ini_pos + _noise_samples - 1, size)
            noise[_ini_pos:_end_pos, :] += np.random.normal(0, 10 ** (noise_events_power_delta_db / 20) * 0.1,
                                                            size=(_end_pos - _ini_pos, n_channels))

    return noise, stationary_noise_var


def generate_line_noise(size: int = None,
                        fs: u.Quantity = None,
                        n_channels: int = None,
                        line_noise_mixing_matrix: np.array = None,
                        line_noise_amplitude: u.Quantity = 10 * u.uV,
                        line_noise_frequency: u.Quantity = 50 * u.Hz,
                        line_noise_n_harmonics: int = 1):
    # generate line noise
    if line_noise_n_harmonics is None:
        line_noise_n_harmonics = int(np.floor((fs / 2) / line_noise_frequency))
    if line_noise_mixing_matrix is None:
        line_noise_mixing_matrix = np.diag(np.ones(n_channels))
    line_coeff = line_noise_mixing_matrix * line_noise_amplitude
    _line_signal = np.zeros((size, 1))
    for _n in range(line_noise_n_harmonics):
        _line_signal[:, 0] = _line_signal[:, 0] + 1 / (_n + 1) * np.sin(
            2.0 * np.pi * line_noise_frequency * (_n + 1) * np.arange(0, size) * u.rad / fs)

    line_signal = np.tile(_line_signal, (1, n_channels)).dot(line_coeff)
    return line_signal



"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools import eeg_epoch_operators as eop
from peegy.processing.tools.template_generator.auditory_waveforms import abr
from peegy.tools.signal_generator.noise_functions import generate_modulated_noise
import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
# Enable below for interactive backend
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

# create synthetic data
fs = 8000.0 * u.Hz
duration = 16 * u.ms
n_channels = 1
n_trials = 1200
noise_dim = 1  # dimensionality of noise
source, _ = abr(fs=fs, time_length=duration)
n_samples = source.shape[0]
coeff = [1]
block_size = 100
s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, n_trials))

desired_snr = 80.0
ini_std = 10.0 ** (-desired_snr / 20.0) * s_std * n_trials ** 0.5
theoretical_rn = ini_std / n_trials ** 0.5

noise = generate_modulated_noise(fs=fs.value,
                                 duration=source.shape[0] * n_trials / fs.value,
                                 f_noise_low=0,
                                 f_noise_high=4000,
                                 attenuation=0,
                                 n_channels=n_channels) * u.uV

noise = ini_std * noise / np.std(noise, axis=0)
noise = eop.et_fold(noise, n_samples)

data = noise + s * (1 + 2 * np.random.random((1, 1, s.shape[2])))
data = 100 * np.reshape(data, [-1, 1], 'F').squeeze().value
x_corr = np.zeros(data.shape)
for _i in np.arange(data.shape[0] - source.shape[0]):
    _ini = _i
    _end = np.minimum(_i + source.shape[0], data.shape[0])
    x1 = data[_ini: _end]
    x2 = source.squeeze().value
    x_corr[_i] = np.sum(x1 * x2) / np.sqrt(np.sum(x1 ** 2) * np.sum(x2 ** 2))
constructed = np.convolve(source.squeeze(), x_corr)
plt.plot(constructed / constructed.max())
plt.plot(data / data.max())



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.epochs_processing_tools import et_mean, et_frequency_mean2, et_snr_in_rois
from peegy.processing.tools.detection.definitions import TimeROI, TimeROIValue
from peegy.definitions.eeg_definitions import EegPeak
from peegy.definitions.channel_definitions import Domain
from peegy.definitions.events import SingleEvent, Events
from peegy.processing.statistics.definitions import FmpTest, TestType
from peegy.definitions.tables import Tables
from peegy.processing.statistics.eeg_statistic_tools import hotelling_t_square_test, f_test
from peegy.processing.tools.eeg_epoch_operators import w_mean, effective_sampling_size
import numpy as np
import pandas as pd
import os
import pyfftw
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
from PyQt5.QtCore import QLibraryInfo
from numba import njit
from scipy.signal import windows
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class EpochData(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 pre_stimulus_interval: u.quantity.Quantity = None,
                 post_stimulus_interval: u.quantity.Quantity = None,
                 baseline_correction_ini_time: u.quantity.Quantity = None,
                 baseline_correction_end_time: u.quantity.Quantity = None,
                 event_code: float = None,
                 base_line_correction: bool = False,
                 at: u.Quantity = None,
                 demean: bool = True,
                 events_mask: int = None,
                 use_window: bool = False,
                 window_type: str = 'tukey',
                 window_arguments: tuple = (0.1, ),
                 **kwargs):
        """
        This class will take a n*m matrix into a k*m*p, where p (number of trials) is determined by the number of
        triggers used to split the data
        :param input_process: InputOutputProcess Class
        :param pre_stimulus_interval: the length (in sec) of the data to be read before the trigger
        :param post_stimulus_interval: the length (in sec) of the data to be read after the trigger
        :param event_code: integer indicating the event code to be used to epoch the data
        :param base_line_correction: whether to perform baseline correction or not. The initial and final time points
        used to compute the baseline correction are given by baseline_correction_ini_time and
        baseline_correction_end_time
        :param baseline_correction_ini_time: initial time (relative to trigger) to compute the mean for baseline
        correction. For example, if trigger is at zero, and you want to compute the baseline correction from -200 ms you
        can set baseline_correction_ini_time = -200 * u.ms
        :param baseline_correction_end_time: end time (relative to trigger) to compute the mean for baseline
        correction. For example, if trigger is at zero, and you want to compute the baseline correction
        between -200 * ms and -50 ms, you can set baseline_correction_ini_time = -200 * u.ms and
        baseline_correction_end_time = -50 * u.ms
        :param at: an array of with times to use as event marks. When at is not empty, these time stamps will be used
        to epoch the data. Event codes from the input_process will be ignored.
        :param demean: whether to remove the mean from each epoch
        :param events_mask: integer value used to masker triggers codes. This is useful to ignore triggers inputs above
        a particular value. For example, if only the first 8 trigger inputs were used (max decimal value is 255), in a
        system with 16 trigger inputs, then the masker could be set to 255 to ignore any trigger from trigger inputs 9
        to 16
        :param use_window: If True a scipy window will be used.
        :param window_type: string indicating the window that will be applied
        :param window_arguments: a tuple with additional parameters to be passed to windows.get_window function
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(EpochData, self).__init__(input_process=input_process, **kwargs)
        self.pre_stimulus_interval = set_default_unit(pre_stimulus_interval, u.s)
        self.post_stimulus_interval = set_default_unit(post_stimulus_interval, u.s)
        self.event_code = event_code
        self.base_line_correction = base_line_correction
        self.demean = demean
        self.events_mask = events_mask
        self.baseline_correction_ini_time = baseline_correction_ini_time
        self.baseline_correction_end_time = baseline_correction_end_time
        self.at = at
        self.use_window = use_window
        self.window_type = window_type
        self.window_arguments = window_arguments

    def transform_data(self):
        if self.event_code is None:
            print('no event code was provided')
            return
        data = self.input_node.data
        baseline_correction_ini_time = 0.0 * u.s if self.baseline_correction_ini_time is None else \
            self.baseline_correction_ini_time
        pre_stimulus_interval = 0.0 * u.s if self.pre_stimulus_interval is None else self.pre_stimulus_interval
        self.input_node.events.mask = self.events_mask
        if self.at is None:
            events_index = self.input_node.events.get_events_index(code=self.event_code,
                                                                   fs=self.input_node.fs)
        else:
            events_index = self.input_node.x_to_samples(self.at)
        post_stimulus_interval = self.post_stimulus_interval
        baseline_correction_end_time = 0.0 * u.s if self.baseline_correction_end_time is None else \
            self.baseline_correction_end_time
        if self.post_stimulus_interval is None:
            post_stimulus_interval = np.percentile(np.diff(events_index), 90) / self.input_node.fs

        # ensure sufficient data for baseline-correction
        if baseline_correction_ini_time < 0:
            pre_stimulus_interval = np.maximum(pre_stimulus_interval, np.abs(baseline_correction_ini_time))
        if baseline_correction_end_time > 0:
            post_stimulus_interval = np.maximum(post_stimulus_interval, baseline_correction_end_time)

        buffer_size = int((post_stimulus_interval + pre_stimulus_interval) * self.input_node.fs)
        events_index = events_index - int((pre_stimulus_interval - self.input_node.x_offset) * self.input_node.fs)
        total_indexes = events_index.size
        # ensure positive index
        _idx_to_keep = np.logical_and(events_index >= 0,
                                      events_index + buffer_size <= self.input_node.data.shape[0])
        events_index = events_index[_idx_to_keep]
        possible_indexes = events_index.size
        if possible_indexes < total_indexes:
            print("{:} events ignored as data length won't fit epoch length".format(
                total_indexes - possible_indexes))
        print('There are {:} events with code {:}'.format(events_index.shape[0], self.event_code))
        epochs = extract_epochs(data=data, buffer_size=buffer_size, events_index=events_index)
        epochs = epochs * data.unit
        print('A total of {:} epochs were obtained using event code {:}, each with a duration of {:.3f}'.format(
            epochs.shape[2],
            self.event_code,
            (buffer_size / self.input_node.fs).to(u.s)))
        if self.demean:
            epochs = epochs - np.mean(epochs, axis=0)
        if self.base_line_correction:
            _ini_sample = max(0, int((pre_stimulus_interval + baseline_correction_ini_time) * self.input_node.fs))
            _end_sample = max(_ini_sample + 1,
                              int((baseline_correction_end_time - baseline_correction_ini_time) * self.input_node.fs) +
                              _ini_sample)
            epochs = epochs - np.mean(epochs[_ini_sample:_end_sample, :, :], axis=0)
        if self.use_window:
            window = windows.get_window((self.window_type, *self.window_arguments), epochs.shape[0])[:, None, None]
            epochs = epochs * window
        self.output_node.data = epochs
        self.output_node.x_offset = pre_stimulus_interval
        self.output_node.events = None


@njit
def extract_epochs(data: np.array = None,
                   events_index: np.array = None,
                   buffer_size: int = None):
    epochs = np.zeros((buffer_size, data.shape[1], events_index.size), dtype=np.float32)
    for i, _event in enumerate(events_index):
        epochs[:, :, i] = data[_event:_event + buffer_size, :]
    return epochs


class RejectEpochs(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 std_threshold: float = None,
                 rejection_percentage: float = None,
                 rejection_threshold: u.quantity.Quantity = None,
                 max_percentage_epochs_above_threshold: float = 1.0,
                 **kwargs):
        """
        This class will epochs where a given threshold has been exceeded. If any channel exceeds the threshold at a
        given trial, that particular epoch will be removed from all channels. This is done to keep the data in a single
        matrix.
        :param input_process: InputOutputProcess Class
        :param std_threshold: float indicating the threshold in terms of standard deviations to remove epochs
        :param max_percentage_epochs_above_threshold: if a single channel has more than max_epochs_above_threshold
        (percentage) epochs, the channel will be removed.
        :param rejection_percentage indicates the percentage of epochs to remove.
        above the threshold, the channel will be removed.
        :param rejection_threshold: indicates level above which epochs will be rejected
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RejectEpochs, self).__init__(input_process=input_process, **kwargs)
        self.std_threshold = set_default_unit(std_threshold, u.dimensionless_unscaled)
        self.rejection_threshold = set_default_unit(rejection_threshold, u.uV)
        self.max_percentage_epochs_above_threshold = max_percentage_epochs_above_threshold
        self.rejection_percentage = rejection_percentage

    def transform_data(self):
        _data = self.input_node.data
        _idx_to_keep = np.ones((_data.shape[2], ), dtype=bool)
        # first we check if a channel needs to be removed  based on max_percentage_epochs_above_threshold
        if self.max_percentage_epochs_above_threshold < 1.0:
            _max_amp = np.max(np.abs(_data), axis=0)
            _epoch_limit = int(self.max_percentage_epochs_above_threshold * _data.shape[2])
            _n_noisy_epochs_per_channel = np.sum(_max_amp > self.rejection_threshold, axis=1)
            _ch_to_remove = np.argwhere(_n_noisy_epochs_per_channel > _epoch_limit).flatten()
            print('Removing a total of {:} channels with more than {:} epochs above the threshold {:}'.format(
                np.sum(_ch_to_remove.size), _epoch_limit, self.rejection_threshold))
            # we assign data to output node to perform a clean delete of bad channels
            self.output_node.data = _data
            self.output_node.delete_channel_by_idx(_ch_to_remove)
            _data = self.output_node.data

        # now using the new data we find and remove the x % of ephocs with the highest peak
        if self.rejection_percentage is not None:
            _max_amp = np.max(np.abs(_data), axis=0)
            rejection_thr = np.quantile(np.max(_max_amp, axis=0), 1 - self.rejection_percentage)
            to_keep = np.max(_max_amp, axis=0, keepdims=True) < rejection_thr
            _idx_to_keep = np.logical_and(_idx_to_keep, np.all(to_keep, axis=0).flatten())
            print('{:} epochs out of {:} will be kept based on a {:} percentage rejection'.format(
                np.all(to_keep, axis=0).sum(),
                _data.shape[2],
                self.rejection_percentage))

        if self.rejection_threshold is not None:
            # now we find epochs indexes across channels which are above threshold
            _max_amp = np.max(np.abs(_data), axis=0)
            to_keep = _max_amp <= self.rejection_threshold
            _idx_to_keep = np.logical_and(_idx_to_keep, np.all(to_keep, axis=0).flatten())
            print('{:} epochs out of {:} will be kept based on a {:} rejection threshold'.format(
                np.all(to_keep, axis=0).sum(),
                _data.shape[2],
                self.rejection_threshold))

        if self.std_threshold is not None:
            _epochs_std = np.std(_data, axis=0)
            _n_noisy_epochs_across_channels = np.sum(np.greater(_epochs_std,
                                                                (np.mean(_epochs_std, 1) +
                                                                 self.std_threshold * np.std(_epochs_std, 1)).reshape(
                                                                    -1, 1)),
                                                     axis=0,
                                                     keepdims=True)
            to_keep = _n_noisy_epochs_across_channels == 0
            _n_std_to_remove = np.sum(~to_keep)
            _std_percentage = 100 * _n_std_to_remove / _data.shape[2]
            _idx_to_keep = np.logical_and(_idx_to_keep, np.all(to_keep, axis=0).flatten())
            print('Rejecting a total of {:} epochs, corresponding to {:.2f}% above {:} std'.format(
                np.sum(~_idx_to_keep), _std_percentage, self.std_threshold))
        # print some information and reject epochs
        _p_rejected = 100 * (1 - np.sum(_idx_to_keep) / _data.shape[2])
        print('Percentage of total epochs rejected {:.2f}%'.format(_p_rejected))
        if _p_rejected == 100:
            print('All epochs were rejected!! check rejection level '
                  '(currently {:}) and run again'.format(self.rejection_percentage))

        self.output_node.data = _data[:, :, _idx_to_keep.flatten()]


class RemoveBadChannelsEpochsBased(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 threshold: float = 300.0,
                 max_epochs_above_threshold: float = 0.5,
                 **kwargs):
        """
        This class will remove a channel based on epochs data. If any channel has more than max_epochs_above_threshold
        percentage of epochs above a given threshold, this channel will be removed.
        :param input_process: InputOutputProcess Class
        :param threshold: the threshold to reject channels with too much noise
        :param max_epochs_above_threshold: if a single channel has more than max_epochs_above_threshold (percentage)
        epochs, the channel will be removed.
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(RemoveBadChannelsEpochsBased, self).__init__(input_process=input_process, **kwargs)
        self.threshold = threshold
        self.max_epoch_above_threshold = max_epochs_above_threshold

    def transform_data(self):
        _data = self.input_node.data
        _max_amp = np.max(np.abs(_data), axis=0)
        _epoch_limit = int(self.max_epoch_above_threshold * _data.shape[2])
        _n_noisy_epochs_per_channel = np.sum(_max_amp > self.threshold, axis=1)
        _ch_to_remove = np.argwhere(_n_noisy_epochs_per_channel > _epoch_limit).flatten()
        print('Removing a total of {:} channels with more than {:} epochs above the threshold {:}'.format(
            np.sum(_ch_to_remove.size), _epoch_limit, self.threshold))
        # we assign data to output node to perform a clean delete of bad channels
        self.output_node.data = _data
        self.output_node.delete_channel_by_idx(_ch_to_remove)


class SortEpochs(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 ascending=False,
                 sort_by: str = 'std',
                 **kwargs):
        """
        This class will sort epochs by maximum amplitude across all trials.
        :param input_process: InputOutputProcess Class
        :param sort_by: string indicating the method to sort ephocs, either 'std' for standard deviation or 'amp' to use
        maximum amplitude per epoch.
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(SortEpochs, self).__init__(input_process=input_process, **kwargs)
        self.ascending = ascending
        self.sort_by = sort_by

    def transform_data(self):
        _data = self.input_node.data
        if self.sort_by == 'amp':
            _measured_value = np.max(np.abs(_data), axis=0)
            _idx_sorted = np.argsort(_measured_value, axis=1)
        if self.sort_by == 'std':
            _measured_value = np.std(_data, axis=0)
            _idx_sorted = np.argsort(_measured_value, axis=1)
        sorted_data = np.zeros(_data.shape) * _data.unit
        if self.ascending:
            for _ch in range(_idx_sorted.shape[0]):
                sorted_data[:, _ch, :] = _data[:, _ch, _idx_sorted[_ch, :][::-1]]
        else:
            for _ch in range(_idx_sorted.shape[0]):
                sorted_data[:, _ch, :] = _data[:, _ch, _idx_sorted[_ch, :]]

        print('Sorting epochs by {:}'.format(self.sort_by))
        self.output_node.data = sorted_data


class AverageEpochs(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 weighted_average: bool = True,
                 weight_across_epochs: bool = True,
                 n_tracked_points: int = None,
                 block_size: int = 10,
                 roi_windows: np.array([TimeROI]) = np.array([TimeROI()]),
                 **kwargs):
        """
        This InputOutputProcess average epochs
        :param input_process: InputOutputProcess Class
        :param weighted_average: if True, weighted average will be used
        :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights
        are computed within epoch (1 / variance across time)
        :param n_tracked_points: number of equally spaced points over time used to estimate residual noise and weights
        :param block_size: number of trials that will be stacked together to estimate the residual noise
        :param roi_windows: time windows used to perform some measures (snr, rms)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(AverageEpochs, self).__init__(input_process=input_process, **kwargs)
        self.weighted_average = weighted_average
        self.n_tracked_points = n_tracked_points
        self.block_size = block_size
        self.roi_windows = roi_windows
        self.weight_across_epochs = weight_across_epochs

    def transform_data(self):
        block_size = max(self.block_size, 5)
        if self.n_tracked_points is None:
            self.n_tracked_points = self.input_node.data.shape[0]
        samples_distance = int(max(self.input_node.data.shape[0] // self.n_tracked_points, 1))
        w_ave, w, rn, cum_rn, w_fft, n, wb, nb, _ = \
            et_mean(epochs=self.input_node.data,
                    block_size=block_size,
                    samples_distance=samples_distance,
                    weighted=self.weighted_average,
                    weight_across_epochs=self.weight_across_epochs
                    )
        points_per_sweep = int(self.input_node.data.shape[0] / samples_distance)
        # compute noise degrees of freedom based on the effective degrees of freedoms
        dof = effective_sampling_size(w)[0, :]
        rn_dof = dof * points_per_sweep
        self.output_node.data = w_ave
        self.output_node.rn = rn
        self.output_node.cum_rn = cum_rn
        self.output_node.w = w
        self.output_node.n = n
        self.output_node.rn_df = rn_dof
        _stats = self.get_snr_table()
        self.output_node.statistical_tests = Tables(table_name=TestType.f_test_time,
                                                    data=_stats,
                                                    data_source=self.name)
        self.output_node.roi_windows = self.roi_windows

    def get_snr_table(self):
        final_snr, _, f_values = et_snr_in_rois(data_node=self.output_node,
                                                roi_windows=self.roi_windows)
        tests = []
        for _iw, _roi in enumerate(self.roi_windows):
            _samples = _roi.get_roi_samples(data_node=self.output_node)
            _ini = _samples[0]
            _end = _samples[-1]
            _label = None
            if self.roi_windows is not None:
                _label = _roi.label

            f_critic, df_num, df_den, p_value = f_test(
                f_values=f_values[:, _iw],
                df_numerator=5 * np.ones(self.output_node.data.shape[1]),
                df_noise=self.output_node.rn_df,
                alpha=self.output_node.alpha)

            for _idx_ch, _ch in enumerate(self.output_node.layout):
                df_den = self.output_node.rn_df
                tests.append(
                    FmpTest(df_1=df_num[_idx_ch],
                            df_2=df_den[_idx_ch],
                            f=f_values[_idx_ch, _iw],
                            f_critic=f_critic[_idx_ch],
                            p_value=p_value[_idx_ch],
                            rn=self.output_node.rn[_idx_ch],
                            snr=final_snr[_idx_ch, _iw],
                            ini_time=self.output_node.samples_to_x([_ini])[0],
                            end_time=self.output_node.samples_to_x([_end])[0],
                            n_epochs=self.output_node.n,
                            channel=_ch.label,
                            label=_label).__dict__
                )
        out = pd.DataFrame(tests)
        return out


class AverageEpochsFrequencyDomain(InputOutputProcess):
    def __init__(self,
                 input_process=InputOutputProcess,
                 n_tracked_points: int = None,
                 block_size: int = 10,
                 test_frequencies: np.array = None,
                 n_fft: int = None,
                 weighted_average: bool = True,
                 weighted_frequency_domain: bool = False,
                 weight_across_epochs: bool = True,
                 weight_frequencies: np.array = None,
                 delta_frequency: u.Quantity = 5. * u.Hz,
                 power_line_frequency: u.Quantity = 50 * u.Hz,
                 n_jobs: int = 1,
                 **kwargs):
        """
        This InputOutputProcess average epochs in the frequency domain
        :param input_process: InputOutputProcess Class
        :param n_tracked_points: number of equally spaced points over time used to estimate residual noise and weights
        :param block_size: number of trials that will be stack together to estimate the residual noise
        :param test_frequencies: numpy array with frequencies that will be used to compute statistics (Hotelling test)
        :param weighted_average: bool indicating if weighted average is to be used, otherwise standard average is used
        :param weighted_frequency_domain: boll indicating if the weghts are compute in the time or frequency domain
        :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights
        are computed within epoch (1 / variance across time)
        :param weight_frequencies: numpy array with frequencies that will be used to estimate weights when
        weighted_average is activated
        :param delta_frequency: frequency size around each test_frequencies to estimate weights and noise
        :param power_line_frequency: frequency of local power line frequency. This will be used to prevent using
        this frequency or its multiples when performing frequency statistics
        :param n_jobs: number of CPUs to compute fft
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(AverageEpochsFrequencyDomain, self).__init__(input_process=input_process, **kwargs)
        self.n_tracked_points = n_tracked_points
        self.block_size = block_size
        self.test_frequencies = set_default_unit(test_frequencies, u.Hz)
        self.weight_frequencies = set_default_unit(weight_frequencies, u.Hz)
        self.weighted_average = weighted_average
        self.weighted_frequency_domain = weighted_frequency_domain
        self.weight_across_epochs = weight_across_epochs
        self.delta_frequency = set_default_unit(delta_frequency, u.Hz)
        self.power_line_frequency = set_default_unit(power_line_frequency, u.Hz)
        self.n_jobs = n_jobs

    def transform_data(self):
        # average processed data across epochs including frequency average
        if self.weighted_frequency_domain:
            if self.weight_frequencies is None and self.test_frequencies is not None:
                print('No specific frequency provided to estimate weights in the frequency-domain. Pooling across '
                      'test frequencies {:}. Results may be inaccurate if test frequencies contain no '
                      'relevant signal'.format(self.test_frequencies.to_string()))
                self.weight_frequencies = self.test_frequencies

        exact_frequencies = np.array([]) * u.dimensionless_unscaled
        freq_samples = np.array([]) * u.dimensionless_unscaled
        if self.weighted_frequency_domain and self.weighted_frequency_domain and self.weight_frequencies is not None:
            w_ave, snr, rn, by_freq_rn, by_freq_snr, w_fft, w, freq_samples, exact_frequencies = et_frequency_mean2(
                epochs=self.input_node.data,
                fs=self.input_node.fs,
                weighted_average=self.weighted_average,
                test_frequencies=self.test_frequencies,
                delta_frequency=self.delta_frequency,
                block_size=self.block_size,
                power_line_frequency=self.power_line_frequency,
                n_jobs=self.n_jobs
            )
        else:
            if self.n_tracked_points is None:
                self.n_tracked_points = self.input_node.data.shape[0]
            samples_distance = int(max(self.input_node.data.shape[0] // self.n_tracked_points, 1))
            w_ave, w, rn, cum_rn, w_fft, n, wb, nb, snr = \
                et_mean(epochs=self.input_node.data,
                        block_size=self.block_size,
                        weighted=self.weighted_average,
                        weight_across_epochs=self.weight_across_epochs,
                        samples_distance=samples_distance
                        )
            if self.test_frequencies is not None:
                n = w_ave.shape[0]
                freqs = np.arange(0, n) * self.input_node.fs / n
                k = np.array([np.argmin(np.abs(freqs - _f)) for _f in self.test_frequencies])
                exact_frequencies = freqs[k]

            if exact_frequencies.size:
                _fft = pyfftw.builders.rfft(self.input_node.data,
                                            overwrite_input=False,
                                            planner_effort='FFTW_ESTIMATE',
                                            axis=0,
                                            threads=self.n_jobs)()

                scaling_factor = 2 / self.input_node.data.shape[0]
                _fft_samples = _fft * scaling_factor * self.input_node.data.unit
                freq_samples = _fft_samples[k, ...]

        hts = []
        for _idx, _f in enumerate(exact_frequencies):
            _freq_samples = freq_samples[_idx, :, :]
            h_samples = np.vstack((np.real(_freq_samples), np.imag(_freq_samples)))
            h_samples = h_samples.reshape((2, *freq_samples.shape[1::]))
            _spectral_amp = w_mean(epochs=_freq_samples.reshape((1, *freq_samples.shape[1::])),
                                   weights=w[0, :, :])
            _hts = hotelling_t_square_test(samples=h_samples, weights=w, channels=self.input_node.layout)
            for _idx_ch, _ht in enumerate(_hts):
                _ht.frequency_tested = _f.squeeze()
                _spectral_phase = np.angle(_spectral_amp[:, _idx_ch]).squeeze()
                _ht.mean_amplitude = np.abs(_spectral_amp[:, _idx_ch]).squeeze()
                _ht.mean_phase = _spectral_phase
            hts = hts + _hts
            # hotelling_t_square_test
        self.output_node.data = w_fft
        self.output_node.n_fft = self.input_node.data.shape[0]
        self.output_node.domain = Domain.frequency
        self.output_node.rn = rn
        self.output_node.snr = snr
        all_ht2 = [_h.__dict__ for _h in hts]
        _stats = pd.DataFrame(all_ht2)
        self.output_node.statistical_tests = Tables(table_name=TestType.hotelling_t2_freq,
                                                    data=_stats,
                                                    data_source=self.name)
        _stats = self.get_frequency_peaks_as_pandas(pd.DataFrame(all_ht2))
        self.output_node.peaks = _stats
        self.output_node.processing_tables_local = Tables(table_name='peaks_frequency',
                                                          data=_stats,
                                                          data_source=self.name)

    @staticmethod
    def get_frequency_peaks_as_pandas(hotelling_tests: pd.DataFrame = None):
        f_peaks = []
        for _i, _s_ht in hotelling_tests.iterrows():
            _f_peak = EegPeak(
                channel=_s_ht.channel,
                x=_s_ht.frequency_tested,
                rn=_s_ht.rn,
                amp=_s_ht.mean_amplitude,
                amp_snr=_s_ht.snr,
                significant=bool(_s_ht.p_value < 0.05),
                peak_label="{:10.1f}".format(_s_ht.frequency_tested),
                show_label=True,
                positive=True,
                domain=Domain.frequency,
                spectral_phase=_s_ht.mean_phase)
            f_peaks.append(_f_peak.__dict__)
        _data_pd = pd.DataFrame(f_peaks)
        return _data_pd


class DeEpoch(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 demean_edges=True,
                 edge_duration: u.Quantity = None,
                 event_code: int = 1,
                 **kwargs):
        """
        This function will convert a n*m*p matrix into an (n*p) * m matrix
        :param input_process: InputOutputProcess Class
        :param demean_edges: whether data should be demeaned at the edgesbefore converting
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(DeEpoch, self).__init__(input_process=input_process, **kwargs)
        self.demean_edges = demean_edges
        self.edge_duration = set_default_unit(edge_duration, u.s)
        self.event_code = event_code

    def transform_data(self):
        _data = self.input_node.data.copy()
        if self.demean_edges:
            if self.edge_duration is None or self.edge_duration == 0 * u.s:
                _edge_samples = 1
            else:
                _edge_samples = int(self.edge_duration * self.input_node.fs)

            for _i in range(1, _data.shape[2]):
                _previous_edge = np.mean(_data[-_edge_samples::, :, _i - 1], axis=0, keepdims=True)
                _current_edge = np.mean(_data[0:_edge_samples, :, _i], axis=0, keepdims=True)
                _data[:, :, _i] = _data[:, :, _i] - (_current_edge - _previous_edge)

        self.output_node.data = np.reshape(np.transpose(_data, [0, 2, 1]), [-1, _data.shape[1]], order='F')
        events = []
        for _i in range(_data.shape[2]):
            events.append(SingleEvent(code=self.event_code,
                                      time_pos=_i * _data.shape[0] / self.input_node.fs,
                                      dur=0))
        events = Events(events=np.array(events))
        self.output_node.events = events


class ApplyFunctionToTimeROI(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 roi_windows: np.array([TimeROI]) = np.array([TimeROI()]),
                 function_object: object = None,
                 test_name: str = None,
                 **kwargs):
        """
        This InputOutputProcess will apply any function to an array of TimeROI. The function_object should be a function
        with a unique output value. This will be computed for each channel and TimeROI.
        :param input_process: InputOutputProcess Class
        :param roi_windows: time windows used to perform some measures (snr, rms)
        :param function_object: the function to apply to each TimeROI
        :param test_name: string to describe what has been calculated. If empty, the function name will be used
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(ApplyFunctionToTimeROI, self).__init__(input_process=input_process, **kwargs)
        self.roi_windows = roi_windows
        self.function_object = function_object
        self.test_name = test_name

    def transform_data(self):
        roi_samples = [roi.get_roi_samples(data_node=self.input_node) for roi in self.roi_windows]
        tests = []
        for _iw in range(len(roi_samples)):
            _samples = roi_samples[_iw]
            _ini = _samples[0]
            _end = _samples[-1]
            _label = None
            if self.roi_windows is not None:
                _label = self.roi_windows[_iw].label
            if self.test_name is not None:
                test_name = self.test_name
            else:
                test_name = self.function_object.__name__
            _value = self.function_object(self.input_node.data[_samples, ...])
            for _idx_ch, _ch in enumerate(self.input_node.layout):
                tests.append(TimeROIValue(function_value=_value[_idx_ch],
                                          label=_label,
                                          test_name=test_name,
                                          ini_time=self.input_node.samples_to_x([_ini])[0],
                                          end_time=self.input_node.samples_to_x([_end])[0],
                                          channel=_ch.label).__dict__)
        _data_pd = pd.DataFrame(tests)
        self.output_node.statistical_tests = Tables(table_name='time_roi_measurements',
                                                    data=_data_pd,
                                                    data_source=self.name)



"""
.. _tut-abr-sim:

##################################
ABR peak detection (Simulated)
##################################

This example simulates an auditory brainstem response, detect peaks, and save results to a database.

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')

from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.pipe.general import ReferenceData, FilterData, RegressOutEOG, ReSampling
from peegy.processing.pipe.epochs import AverageEpochs, EpochData, RejectEpochs
from peegy.processing.pipe.plot import PlotWaveforms
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import abr
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
from peegy.io.storage.data_storage_reading_tools import sqlite_tables_to_pandas, sqlite_waveforms_to_pandas
from peegy.io.storage.plot_tools import plot_time_frequency_responses
import os
import astropy.units as u
import numpy as np


# %%
# Time windows
# ======================
# Time windows are used to define where to look for different peaks

tw = np.array([
    TimePeakWindow(ini_time=4e-3 * u.s, end_time=6e-3 * u.s, label='V', positive_peak=True),
    TimePeakWindow(ini_ref='V', end_time=12e-3 * u.s, label='N_V', positive_peak=False),
    TimePeakWindow(ini_time=4e-3 * u.s, end_ref='V', label='N_III', positive_peak=False),
    TimePeakWindow(ini_time=3e-3 * u.s, end_ref='N_III', label='III', positive_peak=True),
    TimePeakWindow(ini_time=2e-3 * u.s, end_ref='III', label='N_II', positive_peak=False),
    TimePeakWindow(ini_time=2e-3 * u.s, end_ref='N_II', label='II', positive_peak=True),
    TimePeakWindow(ini_time=1e-3 * u.s, end_ref='II', label='N_I', positive_peak=False),
    TimePeakWindow(ini_time=1e-3 * u.s, end_ref='N_I', label='I', positive_peak=True),
])
# %%
# Peak-to-peak Measures
# ==========================
# Peak-to-peak measures are defined based on the labels of the TimePeaks defined above.


pm = np.array([
    PeakToPeakMeasure(ini_peak='I', end_peak='N_I'),
    PeakToPeakMeasure(ini_peak='II', end_peak='N_II'),
    PeakToPeakMeasure(ini_peak='III', end_peak='N_III'),
    PeakToPeakMeasure(ini_peak='V', end_peak='N_V'),
    PeakToPeakMeasure(ini_peak='III', end_peak='V')])

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures


roi_windows = np.array([TimeROI(ini_time=1e-3 * u.s, end_time=1./47 * u.s, label="abr_snr")])

# %%
# Generating some data
# ============================
# First we generate a target signal, in this example, and auditory brainstem response (ABR) with a target final SNR,
# this is, the expected SNR when all trials or epochs are averaged.


fs = 8192.0 * u.Hz
template_waveform, _ = abr(fs=fs, time_length=0.02 * u.s)
n_channels = 4
event_times = np.arange(0, 240.0, 1./47.0) * u.s
n_trials = event_times.shape[0]
desired_snr_db = 6
snr = 10 ** (desired_snr_db / 10) / n_trials

reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.array([0.001, 1, 1, 1])),
                           eog_mixing_matrix=np.diag(np.array([0, -1, 1, 1])),
                           snr=np.array([0.001 ** 2 * snr, snr, snr, snr]),
                           noise_attenuation=3,
                           include_eog_events=True,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='abr_test')
reader.run()
# %%
# Processing pipe line
# ========================
# First we initialize and populate the pipeline


pipeline = PipePool()

pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['CH_0'],
                                       invert_polarity=False)
pipeline['channel_cleaned'] = RegressOutEOG(pipeline['referenced'],
                                            ref_channel_labels=['EOG1'],
                                            method='template')
pipeline['time_filtered_data'] = FilterData(pipeline['channel_cleaned'],
                                            high_pass=47.0 * u.Hz,
                                            low_pass=None)
pipeline.run()
# %%
# Lets see the output of the filter
# --------------------------------------
# First we generate a target signal, in this example, and auditory brainstem response (ABR).

pipeline['time_filtered_data'].plot(plot_input=True,
                                    plot_output=True,
                                    interactive=False,
                                    show_events=False)

# %%
# Processing pipe continue
# ------------------------------
# First we generate a target signal, in this example, and auditory brainstem response (ABR).


pipeline['time_epochs_filtered'] = EpochData(pipeline['time_filtered_data'],
                                             event_code=1.0,
                                             post_stimulus_interval=20 * u.ms,
                                             base_line_correction=False)
pipeline['time_epochs'] = RejectEpochs(pipeline['time_epochs_filtered'],
                                       rejection_threshold=100 * u.uV,
                                       rejection_percentage=0.1,
                                       std_threshold=3.0)
pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          plot_x_lim=[0, 0.012],
                                                          sf_components=np.arange(0, 1),
                                                          return_figures=True)
pipeline['ht2'] = HotellingT2Test(pipeline['time_epochs'],
                                  roi_windows=roi_windows,
                                  block_time=1e-3 * u.s)

pipeline['ht2dss'] = HotellingT2Test(pipeline['dss_time_epochs'],
                                     roi_windows=roi_windows,
                                     block_time=1e-3 * u.s)

pipeline['time_average'] = AverageEpochs(pipeline['time_epochs'],
                                         roi_windows=roi_windows)
pipeline['time_average_dss'] = AverageEpochs(pipeline['dss_time_epochs'],
                                             roi_windows=roi_windows)
# we up-sample to improve time peak detection
pipeline['up_sampled'] = ReSampling(pipeline['time_average'],
                                    new_sampling_rate=fs * 2)

pipeline['up_sampled_dss'] = ReSampling(pipeline['time_average_dss'],
                                        new_sampling_rate=fs * 2)
pipeline.run()
pipeline['up_sampled_dss'].plot(plot_input=True,
                                plot_output=True,
                                ch_to_plot=['CH_0', 'CH_1', 'CH_2'],
                                interactive=False)
pipeline['up_sampled_peaks'] = PeakDetectionTimeDomain(pipeline['up_sampled'],
                                                       time_peak_windows=tw,
                                                       peak_to_peak_measures=pm)
pipeline['up_sampled_peaks_dss'] = PeakDetectionTimeDomain(pipeline['up_sampled_dss'],
                                                           time_peak_windows=tw,
                                                           peak_to_peak_measures=pm)
pipeline['plotter'] = PlotWaveforms(pipeline['up_sampled_peaks'],
                                    overlay=[pipeline['up_sampled_peaks_dss']],
                                    plot_x_lim=[0, 0.012],
                                    return_figures=True)
pipeline.run()

# now we save our data to a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ABR'}
database_path = reader.input_node.paths.file_directory + os.sep + 'abr_test_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['ht2'],
                                                      pipeline['ht2dss'],
                                                      pipeline['up_sampled_peaks'],
                                                      pipeline['up_sampled_dss']],
                                      include_waveforms=True,
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)

# %%
# Read the saved data and show the tables
# ------------------------------------------
df = sqlite_tables_to_pandas(database_path=database_path,
                             tables=['peaks_time',
                                     'f_test_time',
                                     'amplitudes'])
print(df.peaks_time[['anonymous_name', 'channel', 'peak_label', 'x', 'x_unit']])
print(df.f_test_time[['anonymous_name', 'channel', 'f', 'p_value']])
print(df.amplitudes[['anonymous_name', 'channel', 'amp_label', 'amp', 'amp_unit']])

# Read waveforms from generated database and plot them
# ---------------------------------------------------------
df_waves = sqlite_waveforms_to_pandas(database_path=database_path,
                                      group_factors=['Type', 'data_source', 'channel'],
                                      channels=['CH_1', 'CH_2', 'CH_3'])
df_waves = df_waves.loc[df_waves['data_source'] == 'up_sampled_dss']
fig_out = plot_time_frequency_responses(dataframe=df_waves,
                                        rows_by='Type',
                                        cols_by='channel',
                                        title_by='col')
